# Generated by Django 4.1.3 on 2022-12-17 00:43

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('incident', '0003_rule_action'),
    ]

    operations = [
        migrations.AddField(
            model_name='serverossecalert',
            name='ssh_sig',
            field=models.TextField(default=None, null=True),
        ),
        migrations.AlterField(
            model_name='rulecheck',
            name='parent',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='checks', to='incident.rule'),
        ),
    ]




class RestError(Exception):
    def __init__(self, reason, code=None):
        self.reason = reason
        self.code = code

    def __repr__(self):
        return self.reason


class PermissionDeniedException(RestError):
    def __init__(self, reason="permission denied", code=401):
        self.reason = reason
        self.code = code


class RestValidationError(RestError):
    def __init__(self, reason="rest data is not valid", code=123):
        self.reason = reason
        self.code = code




from medialib import youtube

from django.conf import settings

PresetConfig = {
    'name':		'YouTube',
    'description':	'Upload to YouTube',
    'stage':	101,
    'default_use':	'youtube',
    'applicable_kinds': 'VAS',
    'parameters': {
        'access_token': {
            'description':	'Access Token',
            'required':	True,
            'kind':		'S',
            'configurable':	True,
        },
        'categories_scheme': {
            'description':	'Category scheme XML URL',
            'required':	False,
            'kind':		'S',
            'configurable':	False,
            'value':	'http://gdata.youtube.com/schemas/2007/categories.cat',
        },
        'title': {
            'description':	'Video Title Override',
            'required':	False,
            'kind':		'S',
            'configurable':	True,
        },
        'category': {
            'description':	'Category Name',
            'required':	True,
            'kind':		'C',
            'configurable':	True,
            'choices':	'Film,Autos,Music,Animals,Sports,Travel,Games,Comedy,People,News,Entertainment,Education,Howto,Nonprofit,Tech',
        },
        'description': {
            'description':	'Description Text',
            'required':	False,
            'kind':		'S',
            'configurable':	True,
        },
        'keywords': {
            'description':	'Keywords',
            'required':	False,
            'kind':		'S',
            'configurable':	False,
        },
        'private': {
            'description':	'Set private?',
            'required':	True,
            'kind':		'B',
            'configurable':	True,
        },
    },
}


categories = {
    'Animals': 'Pets & Animals',
    'Autos': 'Autos & Vehicles',
    'Comedy': 'Comedy',
    'Education': 'Education',
    'Entertainment': 'Entertainment',
    'Film': 'Film & Animation',
    'Games': 'Gaming',
    'Howto': 'Howto & Style',
    'Music': 'Music',
    'News': 'News & Politics',
    'Nonprofit': 'Nonprofits & Activism',
    'People': 'People & Blogs',
    'Sports': 'Sports',
    'Tech': 'Science & Technology',
    'Travel': 'Travel & Events',
}


def render(item, renditiondef, params):
    rendition = youtube.uploadMediaItem(params["access_token"], item, renditiondef, params)



import random
from datetime import datetime, timedelta
import time
from rest import settings, UberDict
from ws4redis.redis import getRedisClient
from . import utils
from rest import settings
from rest import datem as date_util

_redis_model = None
# THIS IS IN DAYS
METRICS_AUTO_EXPIRE = settings.get("METRICS_AUTO_EXPIRE", True)
METRICS_EXPIRE_MINUTES_S = settings.get("METRICS_EXPIRE_MINUTES", 1) * 86400
METRICS_EXPIRE_HOURLY_S = settings.get("METRICS_EXPIRE_HOURLY", 7) * 86400
METRICS_EXPIRE_DAILY_S = settings.get("METRICS_EXPIRE_DAILY", 360) * 86400


def get_r():
    global _redis_model
    if not _redis_model:
        _redis_model = R()
    return _redis_model


def metric(slug, num=1, category=None, expire=None, date=None, min_granularity=None, group=None):
    """Create/Increment a metric."""
    get_r().metric(slug, num=num, category=category, expire=expire, date=date, min_granularity=min_granularity, group=group)


def gauge(slug, current_value):
    """Set a value for a Gauge"""
    get_r().gauge(slug, current_value)


def set_metric(slug, value, category=None, expire=None, date=None):
    """Create/Increment a metric."""
    get_r().set_metric(slug, value, category=category, expire=expire, date=date)


def get_metric(slug, granularity=None):
    """get a metric."""
    return get_r().get_metric(slug)


def get_metrics(slugs, since=None, granularity="daily", group=None):
    """Create/Increment a metric."""
    return get_r().get_metric_history_chart_data(slugs, since, granularity, group=group)


def get_category_metrics(category, since=None, granularity="daily"):
    """Create/Increment a metric."""
    r = get_r()
    slugs = [utils.to_string(s) for s in list(r.category_slugs(category))]
    return r.get_metric_history_chart_data(slugs, since, granularity)


def get_slugs(category=None):
    """get slug list"""
    if category:
        return list(get_r().category_slugs(category))
    return list(get_r().metric_slugs())


def get_gauge(slug):
    """get a gauge."""
    return get_r().get_gauge(slug)


def get_gauges(slugs):
    """list guages."""
    return get_r().get_gauges(slugs)


def generate_test_metrics(slug='test-metric', num=100, randomize=False,
                          cap=None, increment_value=100):
    """Generate some dummy metrics for the given ``slug``.

    * ``slug`` -- the Metric slug
    * ``num`` -- Number of days worth of metrics (default is 100)
    * ``randomize`` -- Generate random metric values (default is False)
    * ``cap`` -- If given, cap the maximum metric value.
    * ``increment_value`` -- The amount by which we increment metrics on
      subsequent days. If ``randomize`` is True, this value is used to
      generate a ceiling for random values.

    NOTE: This only generates metrics for daily and larger granularities.

    """
    r = get_r()
    i = 0
    if randomize:
        random.seed()

    r.r.sadd(r._metric_slugs_key, slug)  # Store the slug created.
    for date in r._date_range('daily', utils.getLocalDate(datetime.utcnow()) - timedelta(days=num)):
        # Only keep the keys for daily and above granularities.
        keys = r._build_keys(slug, date=date)
        keys = [k for k in keys if k.split(":")[2] not in ['i', 's', 'h']]
        for key in keys:
            # The following is normally done in r.metric, but we're adding
            # metrics for past days here, so this is duplicate code.
            value = i
            if randomize:
                value = random.randint(0, i + increment_value)
            if cap and r.r.get(key) >= cap:
                value = 0  # Dont' increment this one any more.
            r.r.incr(key, value)
        i += increment_value


def delete_test_metrics(slug='test-metric', num=100):
    """Deletes the metrics created by ``generate_test_metrics``."""
    r = get_r()
    for date in r._date_range('daily', utils.getLocalDate(datetime.utcnow()) - timedelta(days=num)):
        keys = r._build_keys(slug, date=date)
        r.r.srem(r._metric_slugs_key, slug)  # remove metric slugs
        r.r.delete(*keys)  # delete the metrics


def dedupe(items):
    """Remove duplicates from a sequence (of hashable items) while maintaining
    order. NOTE: This only works if items in the list are hashable types.

    Taken from the Python Cookbook, 3rd ed. Such a great book!

    """
    seen = set()
    for item in items:
        if item not in seen:
            yield item
            seen.add(item)


class R(object):

    def __init__(self, **kwargs):
        """Creates a connection to Redis, and sets the key used to store a
        set of slugs for all metrics.

        Valid keyword arguments:

        * ``categories_key`` -- The key storing a set of all metric Categories
          (default is "categories")
        * ``metric_slugs_key`` -- The key storing a set of all metrics slugs
          (default is "metric-slugs")
        * ``gauge_slugs_key`` -- The key storing a set of all slugs for gauges
          (default is "gauge-slugs")
        * ``connection_class`` -- class to use to obtain a Redis connection (set in settings.REDIS_METRICS['CONNECTION_CLASS'])
        * ``host`` -- Redis host (set in settings.REDIS_METRICS['HOST'])
        * ``port`` -- Redis port (set in settings.REDIS_METRICS['PORT'])
        * ``db`` -- Redis DB (set in settings.REDIS_METRICS['DB'])
        * ``password`` -- Redis password (set in settings.REDIS_METRICS['PASSWORD'])
        * ``socket_timeout``   -- Redis password (set in
          settings.REDIS_METRICS['SOCKET_TIMEOUT'])
        * ``connection_pool`` -- Redis connection pool info. (set in
          settings.REDIS_METRICS['SOCKET_CONNECTION_POOL'])

        """
        self._categories_key = kwargs.get('categories_key', 'categories')
        self._metric_slugs_key = kwargs.get('metric_slugs_key', 'metric-slugs')
        self._gauge_slugs_key = kwargs.get('gauge_slugs_key', 'gauge-slugs')

        self.r = getRedisClient()

    def categories(self):
        """Returns a set of Categories under which metrics may have been
        organized."""
        return self.r.smembers(self._categories_key)

    def _category_key(self, category):
        return u"c:{0}".format(category)

    def category_slugs(self, category):
        """Returns a set of the metric slugs for the given category"""
        key = self._category_key(category)
        slugs = self.r.smembers(key)
        return slugs

    def _categorize(self, slug, category):
        """Add the ``slug`` to the ``category``. We store category data as
        as set, with a key of the form::

            c:<category name>

        The data is set of metric slugs::

            "slug-a", "slug-b", ...

        """
        key = self._category_key(category)
        self.r.sadd(key, slug)

        # Store all category names in a Redis set, for easy retrieval
        self.r.sadd(self._categories_key, category)

    def metric_slugs(self):
        """Return a set of metric slugs (i.e. those used to create Redis keys)
        for this app."""
        return self.r.smembers(self._metric_slugs_key)

    def metric_slugs_by_category(self):
        """Return a dictionary of metrics data indexed by category:

            {<category_name>: set(<slug1>, <slug2>, ...)}

        """
        result = utils.OrderedDict()
        categories = sorted(self.r.smembers(self._categories_key))
        for category in categories:
            result[category] = self.category_slugs(category)

        # We also need to see the uncategorized metric slugs, so need some way
        # to check which slugs are not already stored.
        categorized_metrics = set([  # Flatten the list of metrics
            slug for sublist in result.values() for slug in sublist
        ])
        f = lambda slug: slug not in categorized_metrics
        uncategorized = list(set(filter(f, self.metric_slugs())))
        if len(uncategorized) > 0:
            result['Uncategorized'] = uncategorized
        return result

    def delete_metric(self, slug):
        """Removes all keys for the given ``slug``."""

        # To remove all keys for a slug, I need to retrieve them all from
        # the set of metric keys, This uses the redis "keys" command, which is
        # inefficient, but this shouldn't be used all that often.
        prefix = "m:{0}:*".format(slug)
        keys = self.r.keys(prefix)
        self.r.delete(*keys)  # Remove the metric data

        # Finally, remove the slug from the set
        self.r.srem(self._metric_slugs_key, slug)

    def set_metric(self, slug, value, category=None, expire=None, date=None):
        """Assigns a specific value to the *current* metric. You can use this
        to start a metric at a value greater than 0 or to reset a metric.

        The given slug will be used to generate Redis keys at the following
        granularities: Seconds, Minutes, Hours, Day, Week, Month, and Year.

        Parameters:

        * ``slug`` -- a unique value to identify the metric; used in
          construction of redis keys (see below).
        * ``value`` -- The value of the metric.
        * ``category`` -- (optional) Assign the metric to a Category (a string)
        * ``expire`` -- (optional) Specify the number of seconds in which the
          metric will expire.
        * ``date`` -- (optional) Specify the timestamp for the metric; default
          used to build the keys will be the current date and time in UTC form.

        Redis keys for each metric (slug) take the form:

            m:<slug>:s:<yyyy-mm-dd-hh-mm-ss> # Second
            m:<slug>:i:<yyyy-mm-dd-hh-mm>    # Minute
            m:<slug>:h:<yyyy-mm-dd-hh>       # Hour
            m:<slug>:<yyyy-mm-dd>            # Day
            m:<slug>:w:<yyyy-num>            # Week (year - week number)
            m:<slug>:m:<yyyy-mm>             # Month
            m:<slug>:y:<yyyy>                # Year

        """
        keys = utils.build_keys(slug, date=date)

        # Add the slug to the set of metric slugs
        self.r.sadd(self._metric_slugs_key, slug)

        # Construct a dictionary of key/values for use with mset
        data = {}
        for k in keys:
            data[k] = value
        self.r.mset(data)

        # Add the category if applicable.
        if category:
            self._categorize(slug, category)

        # Expire the Metric in ``expire`` seconds if applicable.
        if expire:
            for k in keys:
                self.r.expire(k, expire)

    def metric(self, slug, num=1, category=None, expire=None, date=None, min_granularity=None, group=None):
        """Records a metric, creating it if it doesn't exist or incrementing it
        if it does. All metrics are prefixed with 'm', and automatically
        aggregate for Seconds, Minutes, Hours, Day, Week, Month, and Year.

        Parameters:

        * ``slug`` -- a unique value to identify the metric; used in
          construction of redis keys (see below).
        * ``num`` -- Set or Increment the metric by this number; default is 1.
        * ``category`` -- (optional) Assign the metric to a Category (a string)
        * ``expire`` -- (optional) Specify the number of seconds in which the
          metric will expire.
        * ``date`` -- (optional) Specify the timestamp for the metric; default
          used to build the keys will be the current date and time in UTC form.

        Redis keys for each metric (slug) take the form:

            m:<slug>:s:<yyyy-mm-dd-hh-mm-ss> # Second
            m:<slug>:i:<yyyy-mm-dd-hh-mm>    # Minute
            m:<slug>:h:<yyyy-mm-dd-hh>       # Hour
            m:<slug>:<yyyy-mm-dd>            # Day
            m:<slug>:w:<yyyy-num>            # Week (year - week number)
            m:<slug>:m:<yyyy-mm>             # Month
            m:<slug>:y:<yyyy>                # Year

        """
        if date is None:
            date = utils.datetime.utcnow()
        if group:
            date = group.getLocalTime(date)
        elif settings.METRICS_TIMEZONE:
            date = date_util.convertToLocalTime(settings.METRICS_TIMEZONE, date)

        if isinstance(slug, list):
            for s in slug:
                self.metric(s, num, category, expire, date)
            return
        # Add the slug to the set of metric slugs
        self.r.sadd(self._metric_slugs_key, slug)

        if category:
            self._categorize(slug, category)

        # Increment keys. NOTE: current redis-py (2.7.2) doesn't include an
        # incrby method; .incr accepts a second ``amount`` parameter.
        keys = utils.build_keys(slug, date=date, min_granularity=min_granularity)

        # Use a pipeline to speed up incrementing multiple keys
        m_expireat = int(time.time() + METRICS_EXPIRE_MINUTES_S)
        h_expireat = int(time.time() + METRICS_EXPIRE_HOURLY_S)
        d_expireat = int(time.time() + METRICS_EXPIRE_DAILY_S)

        pipe = self.r.pipeline()
        for key in keys:
            res = pipe.incr(key, num)
            if expire:
                pipe.expire(key, expire)
            elif res == 1 and METRICS_AUTO_EXPIRE:
                if ":h:" in key:
                    pipe.expireat(key, h_expireat)
                elif ":i:" in key:
                    pipe.expireat(key, m_expireat)
                elif ":m:" in key or ":y:" in key:
                    continue
                else:
                    pipe.expireat(key, d_expireat)
        pipe.execute()

    def get_metric(self, slug):
        """Get the current values for a metric.

        Returns a dictionary with metric values accumulated for the seconds,
        minutes, hours, day, week, month, and year.

        """
        results = UberDict()
        granularities = utils.granularities()
        keys = utils.build_keys(slug)
        for granularity, key in zip(granularities, keys):
            try:
                results[granularity] = int(self.r.get(key))
            except Exception:
                pass
        return results

    def get_metrics(self, slug_list):
        """Get the metrics for multiple slugs.

        Returns a list of two-tuples containing the metric slug and a
        dictionary like the one returned by ``get_metric``::

            (
                some-metric, {
                    'seconds': 0, 'minutes': 0, 'hours': 0,
                    'day': 0, 'week': 0, 'month': 0, 'year': 0
                }
            )

        """
        # meh. I should have been consistent here, but I'm lazy, so support these
        # value names instead of granularity names, but respect the min/max
        # granularity settings.
        keys = ['seconds', 'minutes', 'hours', 'day', 'week', 'month', 'year']
        key_mapping = {gran: key for gran, key in zip(utils.GRANULARITIES, keys)}
        keys = [key_mapping[gran] for gran in utils.granularities()]

        results = []
        for slug in slug_list:
            metrics = self.r.mget(*utils.build_keys(slug))
            if any(metrics):  # Only if we have data.
                results.append((slug, dict(zip(keys, metrics))))
        return results

    def get_category_metrics(self, category):
        """Get metrics belonging to the given category"""
        slug_list = self.category_slugs(category)
        return self.get_metrics(slug_list)

    def delete_category(self, category):
        """Removes the category from Redis. This doesn't touch the metrics;
        they simply become uncategorized."""
        # Remove mapping of metrics-to-category
        category_key = self._category_key(category)
        self.r.delete(category_key)

        # Remove category from Set
        self.r.srem(self._categories_key, category)

    def reset_category(self, category, metric_slugs):
        """Resets (or creates) a category containing a list of metrics.

        * ``category`` -- A category name
        * ``metric_slugs`` -- a list of all metrics that are members of the
            category.

        """
        key = self._category_key(category)
        if len(metric_slugs) == 0:
            # If there are no metrics, just remove the category
            self.delete_category(category)
        else:
            # Save all the slugs in the category, and save the category name
            self.r.sadd(key, *metric_slugs)
            self.r.sadd(self._categories_key, category)

    def get_metric_history(self, slugs, since=None, to=None, granularity='daily'):
        """Get history for one or more metrics.

        * ``slugs`` -- a slug OR a list of slugs
        * ``since`` -- the date from which we start pulling metrics
        * ``to`` -- the date until which we start pulling metrics
        * ``granularity`` -- seconds, minutes, hourly,
                             daily, weekly, monthly, yearly

        Returns a list of tuples containing the Redis key and the associated
        metric::

            r = R()
            r.get_metric_history('test', granularity='weekly')
            [
                ('m:test:w:2012-52', '15'),
            ]

        To get history for multiple metrics, just provide a list of slugs::

            metrics = ['test', 'other']
            r.get_metric_history(metrics, granularity='weekly')
            [
                ('m:test:w:2012-52', '15'),
                ('m:other:w:2012-52', '42'),
            ]

        """
        if not type(slugs) == list:
            slugs = [slugs]

        # Build the set of Redis keys that we need to get.
        keys = []
        for slug in slugs:
            for date in utils.date_range(granularity, since, to):
                keys += utils.build_keys(slug, date, granularity)
        keys = list(dedupe(keys))

        # Fetch our data, replacing any None-values with zeros
        results = [0 if v is None else int(v) for v in self.r.mget(keys)]
        results = zip(keys, results)
        return sorted(results, key=lambda t: t[0])

    def get_metric_history_as_columns(self, slugs, since=None,
                                      granularity='daily'):
        """Provides the same data as ``get_metric_history``, but in a columnar
        format. If you had the following yearly history, for example::

            [
                ('m:bar:y:2012', '1'),
                ('m:bar:y:2013', '2'),
                ('m:foo:y:2012', '3'),
                ('m:foo:y:2013', '4')
            ]

        this method would provide you with the following data structure::

            [
                ['Period',  'bar',  'foo']
                ['y:2012',  '1',    '3'],
                ['y:2013',  '2',    '4'],
            ]

        Note that this also includes a header column. Data in this format may
        be useful for certain graphing libraries (I'm looking at you Google
        Charts LineChart).

        """
        history = self.get_metric_history(slugs, since, granularity=granularity)
        _history = []  # new, columnar history
        periods = ['Period']  # A separate, single column for the time period
        for s in slugs:
            column = [s]  # story all the data for a single slug
            for key, value in history:
                # ``metric_slug`` extracts the slug from the Redis Key
                if utils.metric_slug(key) == s:
                    column.append(value)

                # Get time period value as first column; This value is
                # duplicated in the Redis key for each value, so this is a bit
                # inefficient, but... oh well.
                period = utils.strip_metric_prefix(key)
                if period not in periods:
                    periods.append(period)

            _history.append(column)  # Remember that slug's column of data

        # Finally, stick the time periods in the first column.
        _history.insert(0, periods)
        return list(zip(*_history))  # Transpose the rows & columns

    def get_metric_history_chart_data(self, slugs, since=None, granularity='daily', group=None):
        """Provides the same data as ``get_metric_history``, but with metrics
        data arranged in a format that's easy to plot with Chart.js. If you had
        the following yearly history, for example::

            [
                ('m:bar:y:2012', '1'),
                ('m:bar:y:2013', '2'),
                ('m:bar:y:2014', '3'),
                ('m:foo:y:2012', '4'),
                ('m:foo:y:2013', '5')
                ('m:foo:y:2014', '6')
            ]

        this method would provide you with the following data structure::

            'periods': ['y:2012', 'y:2013', 'y:2014']
            'data': [
              {
                'slug': 'bar',
                'values': [1, 2, 3]
              },
              {
                'slug': 'foo',
                'values': [4, 5, 6]
              },
            ]

        """
        if since is None:
            since = utils.start_by_granularity(granularity)
        if group:
            since = group.getLocalTime(since)
        elif settings.METRICS_TIMEZONE:
            since = date_util.convertToLocalTime(settings.METRICS_TIMEZONE, since)

        slugs = sorted(slugs)
        history = self.get_metric_history(slugs, since, granularity=granularity)

        # Convert the history into an intermediate data structure organized
        # by periods. Since the history is sorted by key (which includes both
        # the slug and the date, the values should be ordered correctly.
        periods = []
        data = utils.OrderedDict()
        for k, v in history:
            period = utils.slug_to_label(k)
            if period not in periods:
                periods.append(period)

            slug = utils.metric_slug(k)
            if slug not in data:
                data[slug] = []
            data[slug].append(v)

        # Now, reorganize data for our end result.
        metrics = {'periods': periods, 'data': []}
        for slug, values in data.items():
            metrics['data'].append({
                'slug': slug,
                'values': values
            })

        return metrics  # templates still don't like defaultdict's

    # Gauges. Gauges have a different prefix "g:" in order to differentiate
    # them from a metric of the same name.
    def gauge_slugs(self):
        """Return a set of Gauges slugs (i.e. those used to create Redis keys)
        for this app."""
        return self.r.smembers(self._gauge_slugs_key)

    def _gauge_key(self, slug):
        """Make sure our slugs have a consistent format."""
        return "g:{0}".format(utils.slugify(slug))

    def gauge(self, slug, current_value):
        """Set the value for a Gauge.

        * ``slug`` -- the unique identifier (or key) for the Gauge
        * ``current_value`` -- the value that the gauge should display

        """
        k = self._gauge_key(slug)
        self.r.sadd(self._gauge_slugs_key, slug)  # keep track of all Gauges
        self.r.set(k, current_value)

    def get_gauge(self, slug):
        k = self._gauge_key(slug)
        return self.r.get(k)

    def get_gauges(self, slugs):
        output = {}
        for slug in slugs:
            key = self._gauge_key(slug)
            output[slug] = self.r.get(key)
        return output

    def delete_gauge(self, slug):
        """Removes all gauges with the given ``slug``."""
        key = self._gauge_key(slug)
        self.r.delete(key)  # Remove the Gauge
        self.r.srem(self._gauge_slugs_key, slug)  # Remove from the set of keys




# Copyright (C) 2010 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for Google App Engine

Utilities for making it easier to use OAuth 2.0 on Google App Engine.
"""

__author__ = 'jcgregorio@google.com (Joe Gregorio)'

import base64
import cgi
import httplib2
import logging
import os
import pickle
import threading
import time

from google.appengine.api import app_identity
from google.appengine.api import memcache
from google.appengine.api import users
from google.appengine.ext import db
from google.appengine.ext import webapp
from google.appengine.ext.webapp.util import login_required
from google.appengine.ext.webapp.util import run_wsgi_app
from oauth2client import GOOGLE_AUTH_URI
from oauth2client import GOOGLE_REVOKE_URI
from oauth2client import GOOGLE_TOKEN_URI
from oauth2client import clientsecrets
from oauth2client import util
from oauth2client import xsrfutil
from oauth2client.anyjson import simplejson
from oauth2client.client import AccessTokenRefreshError
from oauth2client.client import AssertionCredentials
from oauth2client.client import Credentials
from oauth2client.client import Flow
from oauth2client.client import OAuth2WebServerFlow
from oauth2client.client import Storage

# TODO(dhermes): Resolve import issue.
# This is a temporary fix for a Google internal issue.
try:
  from google.appengine.ext import ndb
except ImportError:
  ndb = None


logger = logging.getLogger(__name__)

OAUTH2CLIENT_NAMESPACE = 'oauth2client#ns'

XSRF_MEMCACHE_ID = 'xsrf_secret_key'


def _safe_html(s):
  """Escape text to make it safe to display.

  Args:
    s: string, The text to escape.

  Returns:
    The escaped text as a string.
  """
  return cgi.escape(s, quote=1).replace("'", '&#39;')


class InvalidClientSecretsError(Exception):
  """The client_secrets.json file is malformed or missing required fields."""


class InvalidXsrfTokenError(Exception):
  """The XSRF token is invalid or expired."""


class SiteXsrfSecretKey(db.Model):
  """Storage for the sites XSRF secret key.

  There will only be one instance stored of this model, the one used for the
  site.
  """
  secret = db.StringProperty()

if ndb is not None:
  class SiteXsrfSecretKeyNDB(ndb.Model):
    """NDB Model for storage for the sites XSRF secret key.

    Since this model uses the same kind as SiteXsrfSecretKey, it can be used
    interchangeably. This simply provides an NDB model for interacting with the
    same data the DB model interacts with.

    There should only be one instance stored of this model, the one used for the
    site.
    """
    secret = ndb.StringProperty()

    @classmethod
    def _get_kind(cls):
      """Return the kind name for this class."""
      return 'SiteXsrfSecretKey'


def _generate_new_xsrf_secret_key():
  """Returns a random XSRF secret key.
  """
  return os.urandom(16).encode("utf-8").hex()


def xsrf_secret_key():
  """Return the secret key for use for XSRF protection.

  If the Site entity does not have a secret key, this method will also create
  one and persist it.

  Returns:
    The secret key.
  """
  secret = memcache.get(XSRF_MEMCACHE_ID, namespace=OAUTH2CLIENT_NAMESPACE)
  if not secret:
    # Load the one and only instance of SiteXsrfSecretKey.
    model = SiteXsrfSecretKey.get_or_insert(key_name='site')
    if not model.secret:
      model.secret = _generate_new_xsrf_secret_key()
      model.put()
    secret = model.secret
    memcache.add(XSRF_MEMCACHE_ID, secret, namespace=OAUTH2CLIENT_NAMESPACE)

  return str(secret)


class AppAssertionCredentials(AssertionCredentials):
  """Credentials object for App Engine Assertion Grants

  This object will allow an App Engine application to identify itself to Google
  and other OAuth 2.0 servers that can verify assertions. It can be used for the
  purpose of accessing data stored under an account assigned to the App Engine
  application itself.

  This credential does not require a flow to instantiate because it represents
  a two legged flow, and therefore has all of the required information to
  generate and refresh its own access tokens.
  """

  @util.positional(2)
  def __init__(self, scope, **kwargs):
    """Constructor for AppAssertionCredentials

    Args:
      scope: string or iterable of strings, scope(s) of the credentials being
        requested.
    """
    self.scope = util.scopes_to_string(scope)

    # Assertion type is no longer used, but still in the parent class signature.
    super(AppAssertionCredentials, self).__init__(None)

  @classmethod
  def from_json(cls, json):
    data = simplejson.loads(json)
    return AppAssertionCredentials(data['scope'])

  def _refresh(self, http_request):
    """Refreshes the access_token.

    Since the underlying App Engine app_identity implementation does its own
    caching we can skip all the storage hoops and just to a refresh using the
    API.

    Args:
      http_request: callable, a callable that matches the method signature of
        httplib2.Http.request, used to make the refresh request.

    Raises:
      AccessTokenRefreshError: When the refresh fails.
    """
    try:
      scopes = self.scope.split()
      (token, _) = app_identity.get_access_token(scopes)
    except app_identity.Error as e:
      raise AccessTokenRefreshError(str(e))
    self.access_token = token


class FlowProperty(db.Property):
  """App Engine datastore Property for Flow.

  Utility property that allows easy storage and retrieval of an
  oauth2client.Flow"""

  # Tell what the user type is.
  data_type = Flow

  # For writing to datastore.
  def get_value_for_datastore(self, model_instance):
    flow = super(FlowProperty,
                 self).get_value_for_datastore(model_instance)
    return db.Blob(pickle.dumps(flow))

  # For reading from datastore.
  def make_value_from_datastore(self, value):
    if value is None:
      return None
    return pickle.loads(value)

  def validate(self, value):
    if value is not None and not isinstance(value, Flow):
      raise db.BadValueError('Property %s must be convertible '
                          'to a FlowThreeLegged instance (%s)' %
                          (self.name, value))
    return super(FlowProperty, self).validate(value)

  def empty(self, value):
    return not value


if ndb is not None:
  class FlowNDBProperty(ndb.PickleProperty):
    """App Engine NDB datastore Property for Flow.

    Serves the same purpose as the DB FlowProperty, but for NDB models. Since
    PickleProperty inherits from BlobProperty, the underlying representation of
    the data in the datastore will be the same as in the DB case.

    Utility property that allows easy storage and retrieval of an
    oauth2client.Flow
    """

    def _validate(self, value):
      """Validates a value as a proper Flow object.

      Args:
        value: A value to be set on the property.

      Raises:
        TypeError if the value is not an instance of Flow.
      """
      logger.info('validate: Got type %s', type(value))
      if value is not None and not isinstance(value, Flow):
        raise TypeError('Property %s must be convertible to a flow '
                        'instance; received: %s.' % (self._name, value))


class CredentialsProperty(db.Property):
  """App Engine datastore Property for Credentials.

  Utility property that allows easy storage and retrieval of
  oath2client.Credentials
  """

  # Tell what the user type is.
  data_type = Credentials

  # For writing to datastore.
  def get_value_for_datastore(self, model_instance):
    logger.info("get: Got type " + str(type(model_instance)))
    cred = super(CredentialsProperty,
                 self).get_value_for_datastore(model_instance)
    if cred is None:
      cred = ''
    else:
      cred = cred.to_json()
    return db.Blob(cred)

  # For reading from datastore.
  def make_value_from_datastore(self, value):
    logger.info("make: Got type " + str(type(value)))
    if value is None:
      return None
    if len(value) == 0:
      return None
    try:
      credentials = Credentials.new_from_json(value)
    except ValueError:
      credentials = None
    return credentials

  def validate(self, value):
    value = super(CredentialsProperty, self).validate(value)
    logger.info("validate: Got type " + str(type(value)))
    if value is not None and not isinstance(value, Credentials):
      raise db.BadValueError('Property %s must be convertible '
                          'to a Credentials instance (%s)' %
                            (self.name, value))
    #if value is not None and not isinstance(value, Credentials):
    #  return None
    return value


if ndb is not None:
  # TODO(dhermes): Turn this into a JsonProperty and overhaul the Credentials
  #                and subclass mechanics to use new_from_dict, to_dict,
  #                from_dict, etc.
  class CredentialsNDBProperty(ndb.BlobProperty):
    """App Engine NDB datastore Property for Credentials.

    Serves the same purpose as the DB CredentialsProperty, but for NDB models.
    Since CredentialsProperty stores data as a blob and this inherits from
    BlobProperty, the data in the datastore will be the same as in the DB case.

    Utility property that allows easy storage and retrieval of Credentials and
    subclasses.
    """
    def _validate(self, value):
      """Validates a value as a proper credentials object.

      Args:
        value: A value to be set on the property.

      Raises:
        TypeError if the value is not an instance of Credentials.
      """
      logger.info('validate: Got type %s', type(value))
      if value is not None and not isinstance(value, Credentials):
        raise TypeError('Property %s must be convertible to a credentials '
                        'instance; received: %s.' % (self._name, value))

    def _to_base_type(self, value):
      """Converts our validated value to a JSON serialized string.

      Args:
        value: A value to be set in the datastore.

      Returns:
        A JSON serialized version of the credential, else '' if value is None.
      """
      if value is None:
        return ''
      else:
        return value.to_json()

    def _from_base_type(self, value):
      """Converts our stored JSON string back to the desired type.

      Args:
        value: A value from the datastore to be converted to the desired type.

      Returns:
        A deserialized Credentials (or subclass) object, else None if the
            value can't be parsed.
      """
      if not value:
        return None
      try:
        # Uses the from_json method of the implied class of value
        credentials = Credentials.new_from_json(value)
      except ValueError:
        credentials = None
      return credentials


class StorageByKeyName(Storage):
  """Store and retrieve a credential to and from the App Engine datastore.

  This Storage helper presumes the Credentials have been stored as a
  CredentialsProperty or CredentialsNDBProperty on a datastore model class, and
  that entities are stored by key_name.
  """

  @util.positional(4)
  def __init__(self, model, key_name, property_name, cache=None, user=None):
    """Constructor for Storage.

    Args:
      model: db.Model or ndb.Model, model class
      key_name: string, key name for the entity that has the credentials
      property_name: string, name of the property that is a CredentialsProperty
        or CredentialsNDBProperty.
      cache: memcache, a write-through cache to put in front of the datastore.
        If the model you are using is an NDB model, using a cache will be
        redundant since the model uses an instance cache and memcache for you.
      user: users.User object, optional. Can be used to grab user ID as a
        key_name if no key name is specified.
    """
    if key_name is None:
      if user is None:
        raise ValueError('StorageByKeyName called with no key name or user.')
      key_name = user.user_id()

    self._model = model
    self._key_name = key_name
    self._property_name = property_name
    self._cache = cache

  def _is_ndb(self):
    """Determine whether the model of the instance is an NDB model.

    Returns:
      Boolean indicating whether or not the model is an NDB or DB model.
    """
    # issubclass will fail if one of the arguments is not a class, only need
    # worry about new-style classes since ndb and db models are new-style
    if isinstance(self._model, type):
      if ndb is not None and issubclass(self._model, ndb.Model):
        return True
      elif issubclass(self._model, db.Model):
        return False

    raise TypeError('Model class not an NDB or DB model: %s.' % (self._model,))

  def _get_entity(self):
    """Retrieve entity from datastore.

    Uses a different model method for db or ndb models.

    Returns:
      Instance of the model corresponding to the current storage object
          and stored using the key name of the storage object.
    """
    if self._is_ndb():
      return self._model.get_by_id(self._key_name)
    else:
      return self._model.get_by_key_name(self._key_name)

  def _delete_entity(self):
    """Delete entity from datastore.

    Attempts to delete using the key_name stored on the object, whether or not
    the given key is in the datastore.
    """
    if self._is_ndb():
      ndb.Key(self._model, self._key_name).delete()
    else:
      entity_key = db.Key.from_path(self._model.kind(), self._key_name)
      db.delete(entity_key)

  def locked_get(self):
    """Retrieve Credential from datastore.

    Returns:
      oauth2client.Credentials
    """
    credentials = None
    if self._cache:
      json = self._cache.get(self._key_name)
      if json:
        credentials = Credentials.new_from_json(json)
    if credentials is None:
      entity = self._get_entity()
      if entity is not None:
        credentials = getattr(entity, self._property_name)
        if self._cache:
          self._cache.set(self._key_name, credentials.to_json())

    if credentials and hasattr(credentials, 'set_store'):
      credentials.set_store(self)
    return credentials

  def locked_put(self, credentials):
    """Write a Credentials to the datastore.

    Args:
      credentials: Credentials, the credentials to store.
    """
    entity = self._model.get_or_insert(self._key_name)
    setattr(entity, self._property_name, credentials)
    entity.put()
    if self._cache:
      self._cache.set(self._key_name, credentials.to_json())

  def locked_delete(self):
    """Delete Credential from datastore."""

    if self._cache:
      self._cache.delete(self._key_name)

    self._delete_entity()


class CredentialsModel(db.Model):
  """Storage for OAuth 2.0 Credentials

  Storage of the model is keyed by the user.user_id().
  """
  credentials = CredentialsProperty()


if ndb is not None:
  class CredentialsNDBModel(ndb.Model):
    """NDB Model for storage of OAuth 2.0 Credentials

    Since this model uses the same kind as CredentialsModel and has a property
    which can serialize and deserialize Credentials correctly, it can be used
    interchangeably with a CredentialsModel to access, insert and delete the
    same entities. This simply provides an NDB model for interacting with the
    same data the DB model interacts with.

    Storage of the model is keyed by the user.user_id().
    """
    credentials = CredentialsNDBProperty()

    @classmethod
    def _get_kind(cls):
      """Return the kind name for this class."""
      return 'CredentialsModel'


def _build_state_value(request_handler, user):
  """Composes the value for the 'state' parameter.

  Packs the current request URI and an XSRF token into an opaque string that
  can be passed to the authentication server via the 'state' parameter.

  Args:
    request_handler: webapp.RequestHandler, The request.
    user: google.appengine.api.users.User, The current user.

  Returns:
    The state value as a string.
  """
  uri = request_handler.request.url
  token = xsrfutil.generate_token(xsrf_secret_key(), user.user_id(),
                                  action_id=str(uri))
  return  uri + ':' + token


def _parse_state_value(state, user):
  """Parse the value of the 'state' parameter.

  Parses the value and validates the XSRF token in the state parameter.

  Args:
    state: string, The value of the state parameter.
    user: google.appengine.api.users.User, The current user.

  Raises:
    InvalidXsrfTokenError: if the XSRF token is invalid.

  Returns:
    The redirect URI.
  """
  uri, token = state.rsplit(':', 1)
  if not xsrfutil.validate_token(xsrf_secret_key(), token, user.user_id(),
                                 action_id=uri):
    raise InvalidXsrfTokenError()

  return uri


class OAuth2Decorator(object):
  """Utility for making OAuth 2.0 easier.

  Instantiate and then use with oauth_required or oauth_aware
  as decorators on webapp.RequestHandler methods.

  Example:

    decorator = OAuth2Decorator(
        client_id='837...ent.com',
        client_secret='Qh...wwI',
        scope='https://www.googleapis.com/auth/plus')


    class MainHandler(webapp.RequestHandler):

      @decorator.oauth_required
      def get(self):
        http = decorator.http()
        # http is authorized with the user's Credentials and can be used
        # in API calls

  """

  def set_credentials(self, credentials):
    self._tls.credentials = credentials

  def get_credentials(self):
    """A thread local Credentials object.

    Returns:
      A client.Credentials object, or None if credentials hasn't been set in
      this thread yet, which may happen when calling has_credentials inside
      oauth_aware.
    """
    return getattr(self._tls, 'credentials', None)

  credentials = property(get_credentials, set_credentials)

  def set_flow(self, flow):
    self._tls.flow = flow

  def get_flow(self):
    """A thread local Flow object.

    Returns:
      A credentials.Flow object, or None if the flow hasn't been set in this
      thread yet, which happens in _create_flow() since Flows are created
      lazily.
    """
    return getattr(self._tls, 'flow', None)

  flow = property(get_flow, set_flow)


  @util.positional(4)
  def __init__(self, client_id, client_secret, scope,
               auth_uri=GOOGLE_AUTH_URI,
               token_uri=GOOGLE_TOKEN_URI,
               revoke_uri=GOOGLE_REVOKE_URI,
               user_agent=None,
               message=None,
               callback_path='/oauth2callback',
               token_response_param=None,
               _storage_class=StorageByKeyName,
               _credentials_class=CredentialsModel,
               _credentials_property_name='credentials',
               **kwargs):

    """Constructor for OAuth2Decorator

    Args:
      client_id: string, client identifier.
      client_secret: string client secret.
      scope: string or iterable of strings, scope(s) of the credentials being
        requested.
      auth_uri: string, URI for authorization endpoint. For convenience
        defaults to Google's endpoints but any OAuth 2.0 provider can be used.
      token_uri: string, URI for token endpoint. For convenience
        defaults to Google's endpoints but any OAuth 2.0 provider can be used.
      revoke_uri: string, URI for revoke endpoint. For convenience
        defaults to Google's endpoints but any OAuth 2.0 provider can be used.
      user_agent: string, User agent of your application, default to None.
      message: Message to display if there are problems with the OAuth 2.0
        configuration. The message may contain HTML and will be presented on the
        web interface for any method that uses the decorator.
      callback_path: string, The absolute path to use as the callback URI. Note
        that this must match up with the URI given when registering the
        application in the APIs Console.
      token_response_param: string. If provided, the full JSON response
        to the access token request will be encoded and included in this query
        parameter in the callback URI. This is useful with providers (e.g.
        wordpress.com) that include extra fields that the client may want.
      _storage_class: "Protected" keyword argument not typically provided to
        this constructor. A storage class to aid in storing a Credentials object
        for a user in the datastore. Defaults to StorageByKeyName.
      _credentials_class: "Protected" keyword argument not typically provided to
        this constructor. A db or ndb Model class to hold credentials. Defaults
        to CredentialsModel.
      _credentials_property_name: "Protected" keyword argument not typically
        provided to this constructor. A string indicating the name of the field
        on the _credentials_class where a Credentials object will be stored.
        Defaults to 'credentials'.
      **kwargs: dict, Keyword arguments are be passed along as kwargs to the
        OAuth2WebServerFlow constructor.
    """
    self._tls = threading.local()
    self.flow = None
    self.credentials = None
    self._client_id = client_id
    self._client_secret = client_secret
    self._scope = util.scopes_to_string(scope)
    self._auth_uri = auth_uri
    self._token_uri = token_uri
    self._revoke_uri = revoke_uri
    self._user_agent = user_agent
    self._kwargs = kwargs
    self._message = message
    self._in_error = False
    self._callback_path = callback_path
    self._token_response_param = token_response_param
    self._storage_class = _storage_class
    self._credentials_class = _credentials_class
    self._credentials_property_name = _credentials_property_name

  def _display_error_message(self, request_handler):
    request_handler.response.out.write('<html><body>')
    request_handler.response.out.write(_safe_html(self._message))
    request_handler.response.out.write('</body></html>')

  def oauth_required(self, method):
    """Decorator that starts the OAuth 2.0 dance.

    Starts the OAuth dance for the logged in user if they haven't already
    granted access for this application.

    Args:
      method: callable, to be decorated method of a webapp.RequestHandler
        instance.
    """

    def check_oauth(request_handler, *args, **kwargs):
      if self._in_error:
        self._display_error_message(request_handler)
        return

      user = users.get_current_user()
      # Don't use @login_decorator as this could be used in a POST request.
      if not user:
        request_handler.redirect(users.create_login_url(
            request_handler.request.uri))
        return

      self._create_flow(request_handler)

      # Store the request URI in 'state' so we can use it later
      self.flow.params['state'] = _build_state_value(request_handler, user)
      self.credentials = self._storage_class(
          self._credentials_class, None,
          self._credentials_property_name, user=user).get()

      if not self.has_credentials():
        return request_handler.redirect(self.authorize_url())
      try:
        resp = method(request_handler, *args, **kwargs)
      except AccessTokenRefreshError:
        return request_handler.redirect(self.authorize_url())
      finally:
        self.credentials = None
      return resp

    return check_oauth

  def _create_flow(self, request_handler):
    """Create the Flow object.

    The Flow is calculated lazily since we don't know where this app is
    running until it receives a request, at which point redirect_uri can be
    calculated and then the Flow object can be constructed.

    Args:
      request_handler: webapp.RequestHandler, the request handler.
    """
    if self.flow is None:
      redirect_uri = request_handler.request.relative_url(
          self._callback_path) # Usually /oauth2callback
      self.flow = OAuth2WebServerFlow(self._client_id, self._client_secret,
                                      self._scope, redirect_uri=redirect_uri,
                                      user_agent=self._user_agent,
                                      auth_uri=self._auth_uri,
                                      token_uri=self._token_uri,
                                      revoke_uri=self._revoke_uri,
                                      **self._kwargs)

  def oauth_aware(self, method):
    """Decorator that sets up for OAuth 2.0 dance, but doesn't do it.

    Does all the setup for the OAuth dance, but doesn't initiate it.
    This decorator is useful if you want to create a page that knows
    whether or not the user has granted access to this application.
    From within a method decorated with @oauth_aware the has_credentials()
    and authorize_url() methods can be called.

    Args:
      method: callable, to be decorated method of a webapp.RequestHandler
        instance.
    """

    def setup_oauth(request_handler, *args, **kwargs):
      if self._in_error:
        self._display_error_message(request_handler)
        return

      user = users.get_current_user()
      # Don't use @login_decorator as this could be used in a POST request.
      if not user:
        request_handler.redirect(users.create_login_url(
            request_handler.request.uri))
        return

      self._create_flow(request_handler)

      self.flow.params['state'] = _build_state_value(request_handler, user)
      self.credentials = self._storage_class(
          self._credentials_class, None,
          self._credentials_property_name, user=user).get()
      try:
        resp = method(request_handler, *args, **kwargs)
      finally:
        self.credentials = None
      return resp
    return setup_oauth


  def has_credentials(self):
    """True if for the logged in user there are valid access Credentials.

    Must only be called from with a webapp.RequestHandler subclassed method
    that had been decorated with either @oauth_required or @oauth_aware.
    """
    return self.credentials is not None and not self.credentials.invalid

  def authorize_url(self):
    """Returns the URL to start the OAuth dance.

    Must only be called from with a webapp.RequestHandler subclassed method
    that had been decorated with either @oauth_required or @oauth_aware.
    """
    url = self.flow.step1_get_authorize_url()
    return str(url)

  def http(self):
    """Returns an authorized http instance.

    Must only be called from within an @oauth_required decorated method, or
    from within an @oauth_aware decorated method where has_credentials()
    returns True.
    """
    return self.credentials.authorize(httplib2.Http())

  @property
  def callback_path(self):
    """The absolute path where the callback will occur.

    Note this is the absolute path, not the absolute URI, that will be
    calculated by the decorator at runtime. See callback_handler() for how this
    should be used.

    Returns:
      The callback path as a string.
    """
    return self._callback_path


  def callback_handler(self):
    """RequestHandler for the OAuth 2.0 redirect callback.

    Usage:
       app = webapp.WSGIApplication([
         ('/index', MyIndexHandler),
         ...,
         (decorator.callback_path, decorator.callback_handler())
       ])

    Returns:
      A webapp.RequestHandler that handles the redirect back from the
      server during the OAuth 2.0 dance.
    """
    decorator = self

    class OAuth2Handler(webapp.RequestHandler):
      """Handler for the redirect_uri of the OAuth 2.0 dance."""

      @login_required
      def get(self):
        error = self.request.get('error')
        if error:
          errormsg = self.request.get('error_description', error)
          self.response.out.write(
              'The authorization request failed: %s' % _safe_html(errormsg))
        else:
          user = users.get_current_user()
          decorator._create_flow(self)
          credentials = decorator.flow.step2_exchange(self.request.params)
          decorator._storage_class(
              decorator._credentials_class, None,
              decorator._credentials_property_name, user=user).put(credentials)
          redirect_uri = _parse_state_value(str(self.request.get('state')),
                                            user)

          if decorator._token_response_param and credentials.token_response:
            resp_json = simplejson.dumps(credentials.token_response)
            redirect_uri = util._add_query_parameter(
                redirect_uri, decorator._token_response_param, resp_json)

          self.redirect(redirect_uri)

    return OAuth2Handler

  def callback_application(self):
    """WSGI application for handling the OAuth 2.0 redirect callback.

    If you need finer grained control use `callback_handler` which returns just
    the webapp.RequestHandler.

    Returns:
      A webapp.WSGIApplication that handles the redirect back from the
      server during the OAuth 2.0 dance.
    """
    return webapp.WSGIApplication([
        (self.callback_path, self.callback_handler())
        ])


class OAuth2DecoratorFromClientSecrets(OAuth2Decorator):
  """An OAuth2Decorator that builds from a clientsecrets file.

  Uses a clientsecrets file as the source for all the information when
  constructing an OAuth2Decorator.

  Example:

    decorator = OAuth2DecoratorFromClientSecrets(
      os.path.join(os.path.dirname(__file__), 'client_secrets.json')
      scope='https://www.googleapis.com/auth/plus')


    class MainHandler(webapp.RequestHandler):

      @decorator.oauth_required
      def get(self):
        http = decorator.http()
        # http is authorized with the user's Credentials and can be used
        # in API calls
  """

  @util.positional(3)
  def __init__(self, filename, scope, message=None, cache=None):
    """Constructor

    Args:
      filename: string, File name of client secrets.
      scope: string or iterable of strings, scope(s) of the credentials being
        requested.
      message: string, A friendly string to display to the user if the
        clientsecrets file is missing or invalid. The message may contain HTML
        and will be presented on the web interface for any method that uses the
        decorator.
      cache: An optional cache service client that implements get() and set()
        methods. See clientsecrets.loadfile() for details.
    """
    client_type, client_info = clientsecrets.loadfile(filename, cache=cache)
    if client_type not in [
        clientsecrets.TYPE_WEB, clientsecrets.TYPE_INSTALLED]:
      raise InvalidClientSecretsError(
          'OAuth2Decorator doesn\'t support this OAuth 2.0 flow.')
    constructor_kwargs = {
      'auth_uri': client_info['auth_uri'],
      'token_uri': client_info['token_uri'],
      'message': message,
    }
    revoke_uri = client_info.get('revoke_uri')
    if revoke_uri is not None:
      constructor_kwargs['revoke_uri'] = revoke_uri
    super(OAuth2DecoratorFromClientSecrets, self).__init__(
        client_info['client_id'], client_info['client_secret'],
        scope, **constructor_kwargs)
    if message is not None:
      self._message = message
    else:
      self._message = 'Please configure your application for OAuth 2.0.'


@util.positional(2)
def oauth2decorator_from_clientsecrets(filename, scope,
                                       message=None, cache=None):
  """Creates an OAuth2Decorator populated from a clientsecrets file.

  Args:
    filename: string, File name of client secrets.
    scope: string or list of strings, scope(s) of the credentials being
      requested.
    message: string, A friendly string to display to the user if the
      clientsecrets file is missing or invalid. The message may contain HTML and
      will be presented on the web interface for any method that uses the
      decorator.
    cache: An optional cache service client that implements get() and set()
      methods. See clientsecrets.loadfile() for details.

  Returns: An OAuth2Decorator

  """
  return OAuth2DecoratorFromClientSecrets(filename, scope,
                                          message=message, cache=cache)



from django.db import models
from django.conf import settings

from rest import models as rm
from rest import helpers as rh


class ServerOssecAlert(models.Model, rm.RestModel):
    # osec alert settings will be managed through metadata
    class RestMeta:
        DEFAULT_SORT = "-id"
        SEARCH_FIELDS = ["rule_id", "hostname", "action"]
        GRAPHS = {
            "default": {
                "graphs": {
                    "geoip": "default"
                }
            },
            "list": {
                "graphs": {
                    "self": "default"
                }
            },
            "detailed": {
                "graphs": {
                    "self": "default"
                }
            }
        }

    created     = models.DateTimeField(db_index=True, auto_now_add=True, editable=False)
    when        = models.DateTimeField(db_index=True)
    alert_id    = models.CharField(max_length=64, blank=False, null=True, default=None)
    rule_id     = models.CharField(db_index=True, max_length=64, blank=False, null=True, default=None)
    text        = models.TextField(null=True, default=None)
    action      = models.CharField(max_length=32, blank=True, null=True, default=None, db_index=True)
    ext_ip      = models.CharField(max_length=32, blank=True, null=True, default=None)
    src_ip      = models.CharField(max_length=32, blank=True, null=True, default=None)
    hostname    = models.CharField(max_length=128, blank=True, null=True, default=None, db_index=True)
    ip          = models.CharField(max_length=32, blank=True, null=True, default=None)
    username    = models.CharField(max_length=64, blank=False, null=True, default=None)
    username2   = models.CharField(max_length=64, blank=False, null=True, default=None)
    ssh_sig     = models.TextField(blank=False, null=True, default=None)

    level       = models.IntegerField(default=0)
    title       = models.CharField(max_length=200, blank=True, null=True, default=None)
    geoip       = models.ForeignKey("location.GeoIP", blank=True, null=True, default=None, on_delete=models.DO_NOTHING)
    
    def __str__(self):
        return f'{self.hostname}: {self.title}'



from objict import objict
from rest import settings
import requests

ABSTRACTAPI_TOKEN = settings.get("GEOIP_ABSTRACTAPI_TOKEN", "817a0affd6b6462189264bc828d0d9f9")


def getLocation(ip):
    """
    {
        "ip_address": "72.219.88.137",
        "city": "Dana Point",
        "city_geoname_id": 5341483,
        "region": "California",
        "region_iso_code": "CA",
        "region_geoname_id": 5332921,
        "postal_code": "92629",
        "country": "United States",
        "country_code": "US",
        "country_geoname_id": 6252001,
        "country_is_eu": false,
        "continent": "North America",
        "continent_code": "NA",
        "continent_geoname_id": 6255149,
        "longitude": -117.7013,
        "latitude": 33.4769,
        "security":
        {
            "is_vpn": false
        },
        "timezone":
        {
            "name": "America/Los_Angeles",
            "abbreviation": "PST",
            "gmt_offset": -8,
            "current_time": "21:55:15",
            "is_dst": false
        },
        "flag":
        {
            "emoji": "🇺🇸",
            "unicode": "U+1F1FA U+1F1F8",
            "png": "https://static.abstractapi.com/country-flags/US_flag.png",
            "svg": "https://static.abstractapi.com/country-flags/US_flag.svg"
        },
        "currency":
        {
            "currency_name": "USD",
            "currency_code": "USD"
        },
        "connection":
        {
            "autonomous_system_number": 22773,
            "autonomous_system_organization": "ASN-CXA-ALL-CCI-22773-RDC",
            "connection_type": "Corporate",
            "isp_name": "Cox Communications Inc.",
            "organization_name": "Cox Communications Inc."
        }
    }
    """
    params = dict(api_key=ABSTRACTAPI_TOKEN)
    resp = requests.get("https://ipgeolocation.abstractapi.com/v1/", params=params)
    data = objict.fromdict(resp.json())

    loc = objict(provider="abstracapi")
    loc.isp = data.isp_name
    loc.ip = data.ip_address
    loc.hostname = ""
    loc.ip_type = ""
    loc.latitude = data.latitude
    loc.longitude = data.longitude
    loc.continent = data.continent
    loc.region = data.region
    loc.country = data.country
    loc.countryCode = data.country_code
    loc.state = data.region_iso_code
    loc.city = data.city
    loc.business = ""
    loc.website = ""
    loc.zipcode = data.postal_code
    return loc




from django.db import models

from rest import models as rm
from rest.extra import JSONMetaData
from rest import log
from rest import settings
from objict import objict

import metrics

from datetime import datetime, timedelta
from .incident import Incident
from .rules import Rule

INCIDENT_METRICS = settings.get("INCIDENT_METRICS", False)
INCIDENT_EVENT_METRICS = settings.get("INCIDENT_EVENT_METRICS", False)
EVENT_TO_INCIDENT_LEVEL = settings.get("EVENT_TO_INCIDENT_LEVEL", 4)


logger = log.getLogger("incident", filename="incident.log")

"""
very generic 
external system can post an event
{
     "description": "Critical Test Event",
     "hostname": "r1",
     "details": "A critical event occurred on r1 running blah blah",
     "level": 7,
     "category": "prepaid.event",
     "metadata": {
        "error_stack": "....."
     }
}
"""


class Event(JSONMetaData, rm.RestModel):
    class RestMeta:
        POST_SAVE_FIELDS = ["level", "catagory"]
        SEARCH_FIELDS = ["description", "hostname"]
        # VIEW_PERMS = ["example_permission"]
        GRAPHS = {
            "default": {
                "graphs": {
                    "group": "basic",
                    "created_by": "basic"
                },
            },
            "detailed": {
                "graphs": {
                    "group": "basic",
                    "created_by": "basic",
                    "generic__component": "basic",
                },
            },
        }

    created = models.DateTimeField(auto_now_add=True)
    reporter_ip = models.CharField(max_length=16, blank=True, null=True, default=None, db_index=True)

    hostname = models.CharField(max_length=255, blank=True, null=True, default=None, db_index=True)
    description = models.CharField(max_length=84)
    details = models.TextField(default=None, null=True)

    level = models.IntegerField(default=0, db_index=True)
    category = models.CharField(max_length=124, db_index=True)
    
    component = models.SlugField(max_length=250, null=True, blank=True, default=None)
    component_id = models.IntegerField(null=True, blank=True, default=None)

    # this allows us to bundle multiple events to an incident
    incident = models.ForeignKey(
        Incident, null=True, default=None, 
        related_name="events", on_delete=models.SET_NULL)

    def runRules(self):
        for rule in Rule.objects.filter(category=self.category).order_by("priority"):
            if rule.run(self):
                return rule
        return None

    def lookupIP(self, ip):
        GeoIP = rm.RestModel.getModel("location", "GeoIP")
        gip = GeoIP.lookup(ip)
        self.setProperty("ip", ip)
        if gip:
            self.setProperty("country", gip.country)
            self.setProperty("state", gip.state)
            self.setProperty("city", gip.city)
            self.setProperty("isp", gip.isp)

    def set_description(self, value):
        # trun desc to 84
        if len(value) > 84:
            value = value[:80] + "..."
        self.description = value

    def on_rest_pre_save(self, request):
        if self.hostname is None:
            self.hostname = settings.HOSTNAME
        if not self.getProperty("hostname", None):
            self.setProperty("hostname", self.hostname)
        self.setProperty("category", self.category)
        if request and request.DATA.get("ip_lookup", field_type=bool):
            self.reporter_ip = request.ip
            self.lookupIP(request.ip)

    def on_rest_saved(self, request, is_new=False):
        if INCIDENT_EVENT_METRICS:
            if self.hostname:
                metrics.metric(f"incident_evt_{self.hostname}", category="incident_events", min_granularity="hourly")
            metrics.metric("incident_evt", category="incident_events", min_granularity="hourly")

        self.setProperty("level", self.level)
        if request is not None:
            self.reporter_ip = request.ip
        # run through rules for the category
        hit_rule = self.runRules()
        priority = 10
        incident = None
        if hit_rule is not None:
            logger.error(f"RULE HIT: {hit_rule.name}")
            priority = hit_rule.priority
            if hit_rule.action == "ignore":
                self.save()
                return
            if hit_rule.bundle > 0:
                incident = Incident.getBundled(rule=hit_rule, event=self)
        elif self.level >= EVENT_TO_INCIDENT_LEVEL:
            # we ignore levels 4 and higher if they did not create a rule
            self.save()
            # logger.info(f"ignore event {self.pk} {self.description}")
            return

        # always create an incident 
        if incident is None:
            incident = Incident(
                rule=hit_rule, priority=priority,
                reporter_ip=self.reporter_ip,
                component=self.category, hostname=self.hostname)
            if hit_rule is not None:
                incident.group = hit_rule.group
            # TODO possibly make this smarter?
            if self.category == "ossec":
                incident.description = f"{self.hostname}: {self.description}"
            else:
                incident.description = self.description
            incident.save()
        self.incident = incident
        self.save()
        if INCIDENT_METRICS:
            if self.hostname:
                metrics.metric(f"incidents_{self.hostname}", category="incidents", min_granularity="hourly")
            metrics.metric("incidents", category="incidents", min_granularity="hourly")

        try:
            incident.triggerAction()
        except Exception:
            logger.exception()




__version_info__ = (0, 3, 0)
__version__ = '.'.join(map(str, __version_info__))

ALL = ['UberDict']

import sys, os
import json
import datetime
import time
from decimal import Decimal

import zlib
import base64

# py2/py3 compatibility
if sys.version_info[0] == 2:
    def iteritems(d):
        return iter(list(d.items()))
else:
    def iteritems(d):
        return list(d.items())


# For internal use only as a value that can be used as a default
# and should never exist in a dict.
_MISSING = object()
# we need these for DJANGO Fields
_MISSING_RAISE_ON_KEYS = ["resolve_expression", "prepare_database_save", "as_sql", "get_placeholder"]

class UberDict(dict):

    """
    A dict that supports attribute-style access and hierarchical keys.

    See `__getitem__` for details of how hierarchical keys are handled,
    and `__getattr__` for details on attribute-style access.

    Subclasses may define a '__missing__' method (must be an instance method
    defined on the class and not just an instance variable) that accepts one
    parameter. If such a method is defined, then a call to `my_UberDict[key]`
    (or the equivalent `my_UberDict.__getitem__(key)`) that fails will call
    the '__missing__' method with the key as the parameter and return the
    result of that call (or raise any exception the call raises).
    """

    def __init__(self, *args, **kwargs):
        """
        Initialize a new `UberDict` using `dict.__init__`.

        When passing in a dict arg, this won't do any special
        handling of values that are dicts. They will remain plain dicts inside
        the `UberDict`. For a recursive init that will convert all
        dict values in a dict to UberDicts, use `UberDict.fromdict`.

        Likewise, dotted keys will not be treated specially, so something
        like `UberDict({'a.b': 'a.b'})` is equivalent to `ud = UberDict()` followed
        by `setattr(ud, 'a.b', 'a.b')`.
        """
        dict.__init__(self, *args, **kwargs)

    def __raise_on_missing__(self, key):
        # you can override this to put keys you want to override on missing or all keys
        return key in _MISSING_RAISE_ON_KEYS

    def __getitem__(self, key):
        """
        Get mapped value for given `key`, or raise `KeyError` if no such
        mapping.

        The `key` may be any value that is valid for a plain `dict`. If the
        `key` is a dotted key (a string like 'a.b' containing one or more
        '.' characters), then the key will be split on '.' and interpreted
        as a sequence of `__getitem__` calls. For example,
        `d.__getitem__('a.b')` would be interpreted as (approximately)
        `d.__getitem__('a').__getitem__('b')`. If the key is not a dotted
        it is treated normally.

        :exceptions:
        - KeyError: if there is no such key on a dict (or object that supports
          `__getitem__`) at any level of the dotted-key traversal.
        - TypeError: if key is not hashable or if an object at some point
          in the dotted-key traversal does not support `__getitem__`.
        """
        if not isinstance(key, str) or '.' not in key:
            return dict.__getitem__(self, key)
        try:
            obj, token = _descend(self, key)
            return _get(obj, token)
        except KeyError:
            # if '__missing__' is defined on the class, then we can delegate
            # to that, but we don't delegate otherwise for consistency with
            # plain 'dict' behavior, which requires '__missing__' to be an
            # instance method and not just an instance variable.
            if hasattr(type(self), '__missing__'):
                return self.__missing__(key)
            raise

    def __setitem__(self, key, value):
        """
        Set `value` for given `key`.

        See `__getitem__` for details of how `key` is intepreted if it is a
        dotted key and for exceptions that may be raised.
        """
        if not isinstance(key, str) or '.' not in key:
            return dict.__setitem__(self, key, value)

        obj, token = _descend(self, key)
        return dict.__setitem__(obj, token, value)

    def __delitem__(self, key):
        """
        Remove mapping for `key` in self.

        See `__getitem__` for details of how `key` is intepreted if it is a
        dotted key and for exceptions that may be raised.
        """
        if not isinstance(key, str) or '.' not in key:
            dict.__delitem__(self, key)
            return
        obj, token = _descend(self, key)
        del obj[token]

    def __getattr__(self, key):
        try:
            # no special treatement for dotted keys, but we need to use
            # 'get' rather than '__getitem__' in order to avoid using
            # '__missing__' if key is not in dict
            val = dict.get(self, key, _MISSING)
            if val is _MISSING:
                if self.__raise_on_missing__(key):
                    raise AttributeError("no attribute '%s'" % (key,))
                return None
            return val
        except KeyError as e:
            raise AttributeError("no attribute '%s'" % (e.args[0],))

    def __setattr__(self, key, value):
        # normal setattr behavior, except we put it in the dict
        # instead of setting an attribute (i.e., dotted keys are
        # treated as plain keys)
        dict.__setitem__(self, key, value)

    def __delattr__(self, key):
        try:
            # no special handling of dotted keys
            dict.__delitem__(self, key)
        except KeyError as e:
            raise AttributeError("no attribute '%s'" % (e.args[0]))

    def __reduce__(self):
        # pickle the contents of a UberDict as a list of items;
        # __getstate__ and __setstate__ aren't needed
        constructor = self.__class__
        instance_args = (list(iteritems(self)),)
        return constructor, instance_args

    def get(self, key, default=None):
        # We can't use self[key] to support `get` here, because a missing key
        # should return the `default` and should not use a `__missing__`
        # method if one is defined (as happens for self[key]).

        # this will grab the first key in the key list and return the value
        if type(key) is list:
            for k in key:
                v = self.get(k, _MISSING)
                if v != _MISSING:
                    return v
            return default

        if not isinstance(key, str) or '.' not in key:
            return dict.get(self, key, default)
        try:
            obj, token = _descend(self, key)
            return _get(obj, token)
        except KeyError:
            return default

    def sort(self, by_value=False, reverse=False):
        if by_value:
            return self.sortByValue(reverse=reverse)
        keys = list(self.sortKeys(reverse=reverse))
        old = self.copy()
        self.clear()
        for key in keys:
            self[key] = old[key]
        return self

    def sortByValue(self, reverse=False):
        marklist = sorted(self.items(), key=lambda x:x[1], reverse=reverse)
        self.clear()
        for key, value in marklist:
            self[key] = value
        return self

    def sortKeys(self, reverse=False):
        return sorted(self.keys(), reverse=reverse)

    def find(self, key, default=None, data=None):
        # this will search the dict for the first key it finds that matches this
        if data is None:
            data = self
        v = data.get(key, _MISSING)
        if v != _MISSING:
            return v
        for k in data:
            d = data.get(k)
            if isinstance(d, dict):
                v = self.find(key, _MISSING, d)
                if v != _MISSING:
                    return v
        return default

    def changes(self, dict2, ignore_keys=None):
        changes = UberDict()
        if not bool(dict2):
            dict2 = UberDict()
        for k in self:
            if ignore_keys and k in ignore_keys:
                continue
            v1 = self.get(k, None)
            v2 = dict2.get(k, None)
            if isinstance(v1, dict):
                if not isinstance(v1, UberDict):
                    v1 = UberDict.fromdict(v1)
                v = v1.changes(v2)
                if len(v):
                    changes[k] = v
            elif isinstance(v1, list) and isinstance(v2, list):
                if bool(set(v1).intersection(set(v2))):
                    changes[k] = v2
            else:
                if v1 != v2:
                    changes[k] = v2
        return changes

    def fromKeys(self, keys):
        # generates a new UberDict, but only with the
        # passed in keys
        d = UberDict()
        for k in keys:
            v = dict.__getitem__(self, k)
            d[k] = v
        return d

    @classmethod
    def fromkeys(self, seq, value=None):
        return UberDict((elem, value) for elem in seq)

    @classmethod
    def fromdict(cls, mapping):
        """
        Create a new `UberDict` from the given `mapping` dict.

        The resulting `UberDict` will be equivalent to the input
        `mapping` dict but with all dict instances (recursively)
        converted to an `UberDict` instance.  If you don't want
        this behavior (i.e., you want sub-dicts to remain plain dicts),
        use `UberDict(mapping)` instead.
        """
        if isinstance(mapping, list):
            out = []
            for lv in mapping:
                out.append(cls.fromdict(lv))
            return out
        elif isinstance(mapping, dict):
            ud = cls()
            for k in mapping:
                v = dict.__getitem__(mapping, k)  # okay for py2/py3
                if isinstance(v, dict):
                    v = cls.fromdict(v)
                elif isinstance(v, list):
                    v = cls.fromdict(v)
                # handle keys like 'key1.key2.key3'
                if '.' in k:
                    skeys = k.split('.')
                    nd = ud
                    while len(skeys) > 1:
                        skey = skeys.pop(0)
                        nd2 = nd.get(skey, cls())
                        dict.__setitem__(nd, skey, nd2)
                        nd = nd2
                    skey = skeys.pop(0)
                    dict.__setitem__(nd, skey, v)
                    continue
                dict.__setitem__(ud, k, v)
            return ud
        elif mapping is None:
            return cls()
        return mapping

    @classmethod
    def fromZIP(cls, data, ignore_errors=True):
        if ignore_errors:
            if not isinstance(data, bytes):
                data = base64.b64decode(data.encode('utf-8'))
            return cls.fromJSON(zlib.decompress(data))
        try:
            if not isinstance(data, bytes):
                data = base64.b64decode(data.encode('utf-8'))
            return cls.fromJSON(zlib.decompress(data))
        except Exception as err:
            pass
        return cls()

    def toZIP(self, as_string=False):
        # compress the dictionary to a zip
        cout = zlib.compress(str.encode(self.toJSON(as_string=True)))
        if as_string:
            return base64.b64encode(cout).decode('utf-8')
        return cout

    def save(self, path):
        with open(path, "w") as f:
            f.write(self.toJSON(as_string=True))
            f.write("\n")

    @classmethod
    def fromFile(cls, path, ignore_errors=False):
        if not os.path.isfile(path) or os.stat(path).st_size == 0:
            f = open(path, "w")
            f.write("{}")
            f.close()
        if not ignore_errors:
            with open(path, "r") as f:
                return cls.fromJSON(f.read())
        try:
            with open(path, "r") as f:
                return cls.fromJSON(f.read())
        except:
            pass
        return cls()

    @classmethod
    def fromJSON(cls, text, ignore_errors=True):
        if not ignore_errors:
            return cls.fromdict(json.loads(text))
        try:
            return cls.fromdict(json.loads(text))
        except Exception:
            pass
        return cls()

    def toJSON(self, as_string=False, fields=None, exclude=None):
        """
        By default this create a json ready dictionary
        or string
        """
        d = dict()
        src = self
        if fields:
            src = self.fromKeys(fields)

        for k in src:
            v = dict.__getitem__(src, k)
            t = type(v)
            if exclude and k in exclude:
                continue
            elif isinstance(v, UberDict):
                d[k] = v.toJSON()
            elif isinstance(v, datetime.date):
                d[k] = time.mktime(v.timetuple())
            elif isinstance(v, datetime.date):
                d[k] = v.strftime("%Y/%m/%d")
            elif isinstance(v, Decimal):
                d[k] = float(v)
            elif hasattr(v, "id"):
                d[k] = v.id
            else:
                d[k] = v
        if as_string:
            return json.dumps(d)
        return d

    def asDict(self, fields=None):
        return self.todict(fields)

    def todict(self, fields=None):
        """
        Create a plain `dict` from this `UberDict`.

        The resulting `dict` will be equivalent to this `UberDict`
        but with every `UberDict` value (recursively) converted to
        a plain `dict` instance.
        """
        d = dict()
        for k in self:
            v = dict.__getitem__(self, k)
            if isinstance(v, UberDict):
                v = v.todict()
            d[k] = v
        return d

    def copy(self):
        """
        Return a shallow copy of this `UberDict`.

        For a deep copy, use `UberDict.fromdict` (as long as there aren't
        plain dict values that you don't want converted to `UberDict`).
        """
        return UberDict(self)

    def extend(self, *args, **kwargs):
        for arg in args:
            if isinstance(arg, dict):
                self.extend(**arg)
        for key in kwargs:
            self[key] = kwargs[key]
        return self

    def setdefault(self, key, default=None):
        """
        If `key` is in the dictionary, return its value.
        If not, insert `key` with a value of `default` and return `default`,
        which defaults to `None`.
        """
        val = self.get(key, _MISSING)
        if val is _MISSING:
            val = default
            self[key] = default
        return val

    def set(self, key, value):
        if "." in key:
            data = self
            tokens = key.split('.')
            for token in tokens[:-1]:
                v = data.get(token)
                if v is None:
                    v = UberDict()
                data[token] = v
                data = v
            data[tokens[-1]] = value
            return None
        self[key] = value
        return None

    def __contains__(self, key):
        return self.get(key, _MISSING) is not _MISSING

    def pop(self, key, *args):
        if not isinstance(key, str) or '.' not in key:
            return dict.pop(self, key, *args)
        try:
            obj, token = _descend(self, key)
        except KeyError:
            if args:
                return args[0]
            raise
        else:
            return dict.pop(obj, token, *args)

    def __dir__(self):
        """
        Expose the expected instance and class attributes and methods
        for the builtin `dir` method, as well as the top-level keys that
        are stored.
        """
        return sorted(set(dir(UberDict)) | set(self.keys()))


# helper to do careful and consistent `obj[name]`
def _get(obj, name):
    """
    Get the indexable value with given `name` from `obj`, which may be
    a `dict` (or subclass) or a non-dict that has a `__getitem__` method.
    """
    try:
        # try to get value using dict's __getitem__ descriptor first
        return dict.__getitem__(obj, name)
    except TypeError:
        # if it's a dict, then preserve the TypeError
        if isinstance(obj, dict):
            raise
        # otherwise try one last time, relying on __getitem__ if any
        return obj[name]


# helper for common use case of traversing a path like 'a.b.c.d'
# to get the 'a.b.c' object and do something to it with the 'd' token
def _descend(obj, key):
    """
    Descend on `obj` by splitting `key` on '.' (`key` must contain at least
    one '.') and using `get` on each token that results from splitting
    to fetch the successive child elements, stopping on the next-to-last.

     A `__getitem__` would do `dict.__getitem__(value, token)` with the
     result, and a `__setitem__` would do `dict.__setitem__(value, token, v)`.

    :returns:
    (value, token) - `value` is the next-to-last object found, and
    `token` is the last token in the `key` (the only one that wasn't consumed
    yet).
    """
    tokens = key.split('.')
    assert len(tokens) > 1
    value = obj
    for token in tokens[:-1]:
        value = _get(value, token)
    return value, tokens[-1]



# -*- coding: utf-8 -*-
import sys
from redis import StrictRedis
from http import client as http_client
from ws4redis import settings as private_settings
from ws4redis.exceptions import WebSocketError, HandshakeError, UpgradeRequiredError, SSLRequiredError
from ws4redis.connection import WebsocketConnection

from django.core.exceptions import PermissionDenied
from django.utils.encoding import force_str
from django import http

from rest.log import getLogger
logger = getLogger("async", filename="async.log")


class WebsocketServerBase(object):
    def __init__(self, redis_connection=None):
        """
        redis_connection can be overriden by a mock object.
        """
        self._websockets = set()  # a list of currently active websockets
        self._redis_connection = redis_connection
        if redis_connection is None:
            self._redis_connection = StrictRedis(**private_settings.WS4REDIS_CONNECTION)

        # clear out all online user connections from redis
        self._redis_connection.delete("users:online:connections")
        self._redis_connection.delete("users:online")
        
    def assure_protocol_requirements(self, environ):
        if environ.get('REQUEST_METHOD') != 'GET':
            raise HandshakeError('HTTP method must be a GET')

        if environ.get('SERVER_PROTOCOL') != 'HTTP/1.1':
            raise HandshakeError('HTTP server protocol must be 1.1')

        if environ.get('HTTP_UPGRADE', '').lower() != 'websocket':
            raise HandshakeError('Client does not wish to upgrade to a websocket')

    @property
    def websockets(self):
        return self._websockets

    def __call__(self, environ, start_response):
        """ Hijack the main loop from the original thread and listen on events on Redis and Websockets"""
        connection = None
        try:
            self.assure_protocol_requirements(environ)
            connection = WebsocketConnection(self, environ, start_response)
            connection.handleComs()
        except WebSocketError:
            logger.exception()
            response = http.HttpResponse(status=1001, content='Websocket Closed')
        except UpgradeRequiredError as excpt:
            logger.exception()
            response = http.HttpResponseBadRequest(status=426, content=excpt)
        except HandshakeError as excpt:
            logger.exception()
            response = http.HttpResponseBadRequest(content=excpt)
        except PermissionDenied as excpt:
            logger.exception("PermissionDenied")
            logger.warning('PermissionDenied: {}'.format(excpt), exc_info=sys.exc_info())
            response = http.HttpResponseForbidden(content=excpt)
        except SSLRequiredError as excpt:
            logger.exception("SSLRequiredError")
            response = http.HttpResponseServerError(content=excpt)
        except Exception as excpt:
            logger.exception()
            response = http.HttpResponseServerError(content=excpt)
        else:
            response = http.HttpResponse()
        finally:
            logger.info("closing websocket")
            if connection:
                connection.release()
            else:
                logger.warning('Starting late response on websocket')
                status_text = http_client.responses.get(response.status_code, 'UNKNOWN STATUS CODE')
                status = '{0} {1}'.format(response.status_code, status_text)
                # headers = list(response._headers.values())
                # if six.PY3:
                #     headers = list(headers)
                start_response(force_str(status), [])
                logger.info('Finish non-websocket response with status code: {}'.format(response.status_code))
        return response




from pushit import utils


def run_update(task):
    proj_info = utils.getProjectInfo(task.data.project, task.data.branch)
    if proj_info is None:
        task.failed(f"{utils.settings.HOSTNAME} - no project for {task.data.project}:{task.data.branch}")
    else:
        utils.updateCode(proj_info.get("updater", None), task.data.branch)
        task.log(f"{utils.settings.HOSTNAME} - completed")
        task.completed()



from django.db import models
from rest import models as rm
from rest import helpers as rh
from .. import geolocate


class Address(models.Model, rm.RestModel):
    class RestMeta:
        GRAPHS = {
            "abstract": {
                "fields":[
                    ('line1', 'street1'),
                    ('line2', 'street2'),
                    'city',
                    'state',
                    ('postalcode', 'zip'),
                    'country',
                ]
            }
        }
    modified_by = models.ForeignKey("account.User", null=True, blank=True, default=None, on_delete=models.CASCADE)
    modified = models.DateTimeField(auto_now=True)
    line1 = models.CharField(max_length=255, blank=True, null=True, default=None)
    line2 = models.CharField(max_length=255, blank=True, null=True, default=None)
    city = models.CharField(max_length=127, blank=True, null=True, default=None)
    state = models.CharField(max_length=127, blank=True, null=True, default=None)
    county = models.CharField(max_length=127, blank=True, null=True, default=None)
    country = models.CharField(max_length=16, blank=True, null=True, default=None)
    postalcode = models.CharField(max_length=32, blank=True, null=True, default=None)
    lat = models.FloatField(default=0.0, blank=True)
    lng = models.FloatField(default=0.0, blank=True)

    def getTimezone(self):
        if self.lat:
            return geolocate.getTimeZone(self.lat, self.lng)
        return None

    def refresh(self):
        try:
            res = geolocate.search("{}, {}, {}".format(self.line1, self.city, self.state))
            self.county = res.county
            self.country = res.short_country
            self.lat = res.latitude
            self.lng = res.longitude
            super(Address, self).save()
        except Exception:
            rh.log_exception("address.refresh")

    def save(self, *args, **kwargs):
        if not self.lat:
            self.refresh()
        super(Address, self).save(*args, **kwargs)



from sessionlog.models import SessionLog

from django.db import models
from django.conf import settings

import copy

from rest.models import RestModel
from rest import helpers
from rest.log import getLogger

import traceback

AUDIT_LOGGER = getLogger("auditlog", filename="auditlog.log")
AUDITLOG_LOGGER_BY_USER = getattr(settings, "AUDITLOG_LOGGER_BY_USER", False)
AUDITLOG_LOGGERS = getattr(settings, "AUDITLOG_LOGGERS", {})

PERSISTENT_LOG_PRINT = getattr(settings, "PERSISTENT_LOG_PRINT", False)


MAX_AUDIT_LINES = 400
MAX_AUDIT_LENGTH = 9000
MAX_AUDIT_LINE_LENGTH = 200


class ConsoleColors:
    BLACK = '\033[30m'
    RED = '\033[31m'
    GREEN = '\033[32m'
    YELLOW = '\033[33m'
    PINK = "\033[35m"
    BLUE = '\033[34m'
    WHITE = '\033[37m'

    HBLACK = '\033[90m'
    HRED = '\033[91m'
    HGREEN = '\033[92m'
    HYELLOW = '\033[93m'
    HBLUE = '\033[94m'
    HPINK = "\033[95m"
    HWHITE = '\033[97m'

    HEADER = '\033[95m'
    FAIL = '\033[91m'
    OFF = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'


LOG_FILTERS = getattr(settings, "AUDIT_LOG_FILTERS", {})


class PersistentLog(models.Model, RestModel):
    class RestMeta:
        DEFAULT_SORT = "-when"
        CAN_SAVE = False
        QUERY_FIELDS = [
            "level", "component", "pkey", "action", 
            "request_method", "request_path", 
            "session__ip", "user",
            "tid", "group", "user__username"]
        SEARCH_FIELDS = ["user__username", "user__last_name", "message", "session__ip", "request_path"]
        SEARCH_TERMS = [
            "message", "action", "component", "tid",
            "pkey", ("component_id", "pkey"),
            ("ip", "session__ip"),
            ("path", "request_path"),
            ("method"), "request_method",
            ("username", "user__username"),
            ("group", "group__pk"),
            ("country", "session__location__country"),
            ("state", "session__location__state"),
            ("country", "session__location__country")]
        DATE_RANGE_FIELD = "when"
        # number of days to go back (increase performance)
        DATE_RANGE_DEFAULT = 7

        FORMATS = {
            "csv": [
                "when",
                "level",
                "request_method",
                "request_path",
                "action",
                "message",
                "session.ip",
                "user.username",
                "group.name",
                "tid"
            ]
        }

        GRAPHS = {
            "default": {
                "graphs": {
                    "user": "default",
                    "group": "default",
                    "session": "default"
                }
            },
            "list": {
                "graphs": {
                    "user": "default",
                    "group": "default",
                    "session": "default"
                }
            }
        }

    when = models.DateTimeField(db_index=True, auto_now_add=True, editable=False, help_text="When change was logged")
    level = models.IntegerField(default=0)
    message = models.TextField(blank=True, null=True, default=None)

    action = models.TextField(max_length=127, blank=True, null=True, default=None, db_index=True)
    component = models.CharField(max_length=127, blank=True, null=True, default=None, db_index=True)
    pkey = models.IntegerField(blank=True, null=True, default=None, db_index=True)
    session = models.ForeignKey(SessionLog, null=True, blank=True, on_delete=models.SET_NULL)
    request_method = models.CharField(max_length=254, blank=True, null=True, default=None)
    tid = models.CharField(max_length=254, blank=True, null=True, default=None, db_index=True)
    request_path = models.TextField(blank=True, null=True, default=None)
    user = models.ForeignKey("account.User", null=True, default=None, on_delete=models.CASCADE)
    group = models.ForeignKey("account.Group", null=True, default=None, on_delete=models.CASCADE)

    def setMessage(self, message, truncate=True, sanatize=True):
        if message is None:
            self.message = "None"
            return
        if isinstance(message, dict) or type(message) is list:
            if sanatize and isinstance(message, dict):
                message = PersistentLog.sanatize(message)
            self.message = helpers.dictToString(message, no_truncate=not truncate)
        else:
            self.message = helpers.toString(message)

        if truncate:
            if self.message.count('\n') > MAX_AUDIT_LINES:
                p = helpers.find_nth(self.message, '\n', MAX_AUDIT_LINES)
                self.message = self.message[:p]
                self.message += "\n ... <truncated lines>"

            if len(self.message) > MAX_AUDIT_LENGTH:
                lines = self.message.split('\n')
                tlines = []
                for ln in lines:
                    if len(ln) >= MAX_AUDIT_LINE_LENGTH:
                        tlines.append(ln[:MAX_AUDIT_LINE_LENGTH] + '...<truncated line len>')
                    else:
                        tlines.append(ln)
                self.message = '\n'.join(tlines)

        # check for invalid characters
        if self.message.find("\0") >= 0:
            self.message = self.message.replace("\0", "<null>")
            # helpers.log_error("found null characters", self.message)

    def printToLog(self):
        print_log = ["{}{}:{}{}".format(
            ConsoleColors.YELLOW, self.request_path, self.request_method, ConsoleColors.OFF)]
        if self.action == "request":
            print_log.append("{0}------- BEGIN --------------".format(ConsoleColors.YELLOW))
        print_log.append('{0}***** BEGIN {1}.{2} *****{3}'.format(
            ConsoleColors.YELLOW, self.component, self.action, ConsoleColors.HGREEN))
        if self.user and self.user.is_authenticated:
            print_log.append("USER: {}".format(self.user.username))
        if self.group:
            print_log.append("GROUP: {}".format(self.group.name))
        if self.tid:
            print_log.append("TID: {}".format(self.tid))
        if self.message:
            print_log.append(helpers.toString(self.message))
        else:
            print_log.append("no data")
        print_log.append('{0}***** END {1}.{2} *****{3}'.format(
            ConsoleColors.YELLOW, self.component, self.action, ConsoleColors.OFF))
        if self.action == "response":
            print_log.append("{}------- END {}:{}  --------------{}\n".format(
                ConsoleColors.YELLOW, self.request_path, self.request_method, ConsoleColors.OFF))
        if print_log:
            helpers.log_print(*print_log)

    @staticmethod
    def sanatize_password(value, n=1):
        if value is None or not isinstance(value, str):
            return value
        st = value
        return ''.join("*" if i % n == 0 else char for i, char in enumerate(st, 1))

    @staticmethod
    def sanatize_track(value):
        if value is None or not isinstance(value, str):
            return value
        st = value
        if len(st) > 12:
            pan = None
            pan2 = None
            parts = st.split('?')
            for p1 in parts:
                p1 = p1.upper().replace(';', '').replace('%B', '').replace('B', '')
                ex_data = None
                if '=' in p1:
                    pos = p1.find('=')
                    pan = p1[:pos]
                    ex_data = p1[pos + 8:]
                    st = st.replace(ex_data, "*" * len(ex_data))
                elif '^' in p1:
                    pan = p1[:p1.find('^')]
                    pos = p1.rfind('^')
                    ex_data = p1[pos + 8:]
                    st = st.replace(ex_data, "*" * len(ex_data))
                elif "D2" in p1:
                    p1 = p1.replace("D2", "=")
                    pos = p1.find('=')
                    pan = p1[:pos]
                    ex_data = p1[pos + 8:]
                    st = st.replace(ex_data, "*" * len(ex_data))
                else:
                    continue
                if not pan2:
                    pan2 = pan[:6] + "***" + pan[-4:]
            if pan is not None:
                return st.replace(pan, pan2)
        return value

    @staticmethod
    def sanatize_pan(value):
        if value and len(value) > 10:
            return "{}***{}".format(value[:6], value[-4:])
        return value

    @staticmethod
    def sanatize_epb(value):
        if isinstance(value, str):
            return "{}:*****".format(len(value))
        return value

    @staticmethod
    def sanatize_ssn(value):
        if value:
            return "***-**-{}".format(value[-4:])
        return value

    @staticmethod
    def sanatize_all(value):
        if isinstance(value, str):
            return "*" * len(value)
        return value

    @staticmethod
    def sanatize(obj):
        # deepcopy vs copy?
        sobj = copy.copy(obj)
        for key in sobj:
            lkey = key.lower()
            if lkey in LOG_FILTERS:
                funcn = LOG_FILTERS.get(lkey)
                if callable(funcn):
                    sobj[key] = funcn(sobj[key])
                elif hasattr(PersistentLog, funcn):
                    sobj[key] = getattr(PersistentLog, funcn)(sobj[key])
            elif lkey in ["pan", "cardnumber"]:
                sobj[key] = PersistentLog.sanatize_pan(sobj[key])
            elif "key" in lkey or lkey in ["5a", "57", "99"]:
                sobj[key] = PersistentLog.sanatize_all(sobj[key])
            elif lkey.startswith("track"):
                sobj[key] = PersistentLog.sanatize_track(sobj[key])
            elif "password" in lkey:
                sobj[key] = PersistentLog.sanatize_password(sobj[key])
            elif lkey == "epb":
                sobj[key] = PersistentLog.sanatize_epb(sobj[key])
            elif lkey == "ssn":
                sobj[key] = PersistentLog.sanatize_ssn(sobj[key])
            # lets truncate
            if type(sobj[key]) in [str, str] and len(sobj[key]) > MAX_AUDIT_LINE_LENGTH:
                sobj[key] = sobj[key][:MAX_AUDIT_LINE_LENGTH] + '...<truncated linelen>'
            if isinstance(sobj[key], dict):
                sobj[key] = PersistentLog.sanatize(sobj[key])
        return sobj

    @staticmethod
    def log(message, level=0, request=None, component=None,
            pkey=None, action=None, group=None, path=None, method=None, tid=None, no_truncate=False):
        plog = PersistentLog.createLogFromRequest(
            request, component=component, tid=tid, pkey=pkey, action=action, group=group,
            path=path, method=method, level=level)

        plog.setMessage(message, truncate=not no_truncate)

        try:
            plog.save()
        except Exception:
            helpers.log_exception(plog.message)

        if PERSISTENT_LOG_PRINT:
            plog.printToLog()
        return plog

    @staticmethod
    def logException(message=None, level=50, request=None, component=None,
                     pkey=None, action="error", group=None, path=None,
                     method=None, tid=None, no_truncate=True):
        # LETS LOG THE REQUEST DATA TO BE SURE
        if request is None:
            from rest.middleware import get_request
            request = get_request()

        helpers.log_exception(message)
        if request:
            helpers.log_error(request.DATA.toDict())
            # request.DATA.log()
            helpers.log_error("tid={}\npath={}\nip={}\nuser={}".format(
                tid, path, request.ip, request.user))

        # LOG THE EXCEPTION TO local LOGGERS
        plog = PersistentLog.createLogFromRequest(
            request=request, component=component, tid=tid, pkey=pkey, action=action, group=group,
            path=path, method=method, level=level)
        stack = str(traceback.format_exc())
        plog.setMessage(stack, truncate=not no_truncate)
        try:
            plog.save()
        except Exception:
            helpers.log_exception(plog.message)
        return plog

    @staticmethod
    def logError(message=None, level=0, request=None, component=None,
                 pkey=None, action="error", group=None, path=None,
                 method=None, tid=None, no_truncate=True):
        # LETS LOG THE REQUEST DATA TO BE SURE
        if request is None:
            from rest.middleware import get_request
            request = get_request()
        if request:
            request.DATA.log()
        plog = PersistentLog.createLogFromRequest(
            request, component=component, tid=tid, pkey=pkey, action=action, group=group,
            path=path, method=method, level=level)
        stack = str(traceback.format_exc())
        plog.setMessage(stack, truncate=not no_truncate)
        try:
            plog.save()
        except Exception:
            helpers.log_exception(plog.message)
        # print all errors to local logs
        plog.printToLog()
        return plog

    @staticmethod
    def createLogFromRequest(request=None, component=None, tid=None,
                             pkey=None, action=None, group=None,
                             path=None, method=None, level=10):
        if request is None:
            from rest.middleware import get_request
            request = get_request()
        if pkey is None and request is not None and hasattr(request, "_log_component"):
            if request._log_pk is not None:
                component = request._log_component
                pkey = request._log_pk
        plog = PersistentLog(
            group=group, level=level, component=component, tid=tid,
            pkey=pkey, action=action, request_path=path, request_method=method)
        if request:
            if request.user.is_authenticated:
                plog.user = request.user
            if not group and hasattr(request, "group"):
                plog.group = request.group
            if not path:
                plog.request_path = request.path
            else:
                plog.request_path = path
            if not method:
                plog.request_method = request.method
            else:
                plog.request_method = method
            plog.session = SessionLog.GetSession(request)
            if plog.tid is None:
                if hasattr(request, "terminal") and request.terminal:
                    plog.tid = request.terminal.tid
        return plog


class AuditLog(models.Model):
    """
    Audit Log
    """

    class Meta:
        permissions = (
            ("can_read", "Can read audit log"),
        )

    when = models.DateTimeField(auto_now_add=True, editable=False, help_text="When change was logged")
    model = models.CharField(max_length=127, db_index=True)
    pkey = models.IntegerField(db_index=True)
    attribute = models.CharField(max_length=127)
    user = models.ForeignKey("account.User", null=True, on_delete=models.CASCADE)
    session = models.ForeignKey(SessionLog, null=True, blank=True, on_delete=models.CASCADE)
    how = models.TextField(max_length=127, null=True, blank=True)
    referer = models.TextField(null=True, blank=True)
    stack = models.TextField(null=True, blank=True)
    oldval = models.TextField(null=True, blank=True)
    newval = models.TextField(null=True, blank=True)

    def __unicode__(self):
        return "[%s] %s.%s" % (self.when, self.model, self.attribute)

    def reference(self):
        if self.referer:
            try:
                ret = self.referer.partition(settings.BASE_URL)
            except AttributeError:
                pass
            else:
                if ret[0] == '':
                    return ret[2]
            try:
                ret = self.referer.partition(settings.BASE_URL_SECURE)
            except AttributeError:
                pass
            else:
                if ret[0] == '':
                    return ret[2]

        try:
            return self.session.user_agent.partition(' ')[0]
        except (SessionLog.DoesNotExist, AttributeError):
            pass

        return None



from rest import log
import os
from functools import wraps
import time

# from line_profiler import LineProfiler
# profiler = LineProfiler()

logger = log.getLogger("profiler", filename="profiler.log")


# def profile(func):
#     def inner(*args, **kwargs):
#         profiler.add_function(func)
#         profiler.enable_by_count()
#         return func(*args, **kwargs)
#     return inner


# def logStats():
#     # profiler.dump_stats(os.path.join(log.LOG_FOLDER, "profiler.log"))
#     profiler.print_stats(stream=logger.stream)


def timeit(func):
    @wraps(func)
    def timeit_wrapper(*args, **kwargs):
        start_time = time.perf_counter()
        result = func(*args, **kwargs)
        end_time = time.perf_counter()
        total_time = end_time - start_time
        logger.info(f'Function {func.__name__} Took {total_time:.4f} seconds')
        # logger.info(f'Function {func.__name__}{args} {kwargs} Took {total_time:.4f} seconds')
        return result
    return timeit_wrapper


#!/usr/bin/python2.5
#
# Copyright 2010 the Melange authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Helper methods for creating & verifying XSRF tokens."""

__authors__ = [
  '"Doug Coker" <dcoker@google.com>',
  '"Joe Gregorio" <jcgregorio@google.com>',
]


import base64
import hmac
import os  # for urandom
import time

from oauth2client import util


# Delimiter character
DELIMITER = ':'

# 1 hour in seconds
DEFAULT_TIMEOUT_SECS = 1*60*60

@util.positional(2)
def generate_token(key, user_id, action_id="", when=None):
  """Generates a URL-safe token for the given user, action, time tuple.

  Args:
    key: secret key to use.
    user_id: the user ID of the authenticated user.
    action_id: a string identifier of the action they requested
      authorization for.
    when: the time in seconds since the epoch at which the user was
      authorized for this action. If not set the current time is used.

  Returns:
    A string XSRF protection token.
  """
  when = when or int(time.time())
  digester = hmac.new(key)
  digester.update(str(user_id))
  digester.update(DELIMITER)
  digester.update(action_id)
  digester.update(DELIMITER)
  digester.update(str(when))
  digest = digester.digest()

  token = base64.urlsafe_b64encode('%s%s%d' % (digest,
                                               DELIMITER,
                                               when))
  return token


@util.positional(3)
def validate_token(key, token, user_id, action_id="", current_time=None):
  """Validates that the given token authorizes the user for the action.

  Tokens are invalid if the time of issue is too old or if the token
  does not match what generateToken outputs (i.e. the token was forged).

  Args:
    key: secret key to use.
    token: a string of the token generated by generateToken.
    user_id: the user ID of the authenticated user.
    action_id: a string identifier of the action they requested
      authorization for.

  Returns:
    A boolean - True if the user is authorized for the action, False
    otherwise.
  """
  if not token:
    return False
  try:
    decoded = base64.urlsafe_b64decode(str(token))
    token_time = int(decoded.split(DELIMITER)[-1])
  except (TypeError, ValueError):
    return False
  if current_time is None:
    current_time = time.time()
  # If the token is too old it's not valid.
  if current_time - token_time > DEFAULT_TIMEOUT_SECS:
    return False

  # The given token should match the generated one with the same time.
  expected_token = generate_token(key, user_id, action_id=action_id,
                                  when=token_time)
  if len(token) != len(expected_token):
    return False

  # Perform constant time comparison to avoid timing attacks
  different = 0
  for x, y in zip(token, expected_token):
    different |= ord(x) ^ ord(y)
  if different:
    return False

  return True



"""
A command to record system metrics (CUP, Memory, Disk usage). Run this as a
cron job to periodically caputes sysem resource usage. Requires psutil.

Usage:

    manage.py system_metric <metric> [device|interface] [category]

Examples:

    manage.py system_metric cpu -- Records total CPU usage as a percent
    manage.py system_metric mem -- Records virtual memory used as a percent
    manage.py system_metric disk sda1 -- Records percent usage of device sda1
    manage.py system_metric net eth0 -- Records bytest sent/received on
        interface eth0

"""
from __future__ import unicode_literals
from django.core.management.base import BaseCommand, CommandError
from redis_metrics.utils import set_metric, metric, gauge

try:
    import psutil
except ImportError:
    psutil = None  # NOQA


class Command(BaseCommand):
    args = '<metric_name> [dev|interface] [category]'
    help = """Record System Metrics. Valid metric_name values: cpu, mem, disk.

    USAGE:

    To record CPU usage (percent used), run:

        manage.py system_metric cpu

    To record Memory usage (virtual memory, percent used), run:

        manage.py system_metric mem

    To record Disk usage for a device, run:

        manage.py system_metric disk <device>

    for example:

        mange.py system_metric disk sda0

    To record Network usage for an interface, run:

        manage.py system_metric disk <device>

    for example:

        mange.py system_metric net eth0


    """

    def process_args(self, *args):

        self.category = "System"
        self.metric_name = None
        self.device = None

        if len(args) == 1 and args[0] in ['cpu', 'mem']:
            self.metric_name = args[0]
        elif len(args) == 2 and args[0] in ['cpu', 'mem']:
            self.metric_name = args[0]
            self.category = args[1].strip()
        elif len(args) == 2 and args[0] in ['disk', 'net']:
            self.metric_name = args[0]
            self.device = args[1]
        elif len(args) == 3 and args[0] in ['disk', 'net']:
            self.metric_name = args[0]
            self.device = args[1]
            self.category = args[2].strip()
        else:
            raise CommandError("Invalid Arugments. Plase run "
                "`manage.py help system_metric`")

    def _cpu(self):
        """Record CPU usage."""
        value = int(psutil.cpu_percent())
        set_metric("cpu", value, category=self.category)
        gauge("cpu", value)

    def _mem(self):
        """Record Memory usage."""
        value = int(psutil.virtual_memory().percent)
        set_metric("memory", value, category=self.category)
        gauge("memory", value)

    def _disk(self):
        """Record Disk usage."""
        mountpoints = [
            p.mountpoint for p in psutil.disk_partitions()
            if p.device.endswith(self.device)
        ]
        if len(mountpoints) != 1:
            raise CommandError("Unknown device: {0}".format(self.device))

        value = int(psutil.disk_usage(mountpoints[0]).percent)
        set_metric("disk-{0}".format(self.device), value, category=self.category)
        gauge("disk-{0}".format(self.device), value)

    def _net(self):
        """Record Network usage."""
        data = psutil.network_io_counters(pernic=True)
        if self.device not in data:
            raise CommandError("Unknown device: {0}".format(self.device))

        # Network bytes sent
        value = data[self.device].bytes_sent
        metric("net-{0}-sent".format(self.device), value, category=self.category)
        gauge("net-{0}-sent".format(self.device), value)

        # Network bytes received
        value = data[self.device].bytes_recv
        metric("net-{0}-recv".format(self.device), value, category=self.category)

    def handle(self, *args, **options):
        # Make sure we've got psutil
        if psutil is None:
            raise CommandError("This command requires psutil")

        self.process_args(*args)

        if self.metric_name == "cpu":
            self._cpu()
        elif self.metric_name == "mem":
            self._mem()
        elif self.metric_name == "disk":
            self._disk()
        elif self.metric_name == "net":
            self._net()
        else:
            m = "{0} is an invalid metric_name".format(self.metric_name)
            raise CommandError(m)



# Generated by Django 4.0.2 on 2022-02-22 01:25

from django.db import migrations, models
import django.db.models.deletion
import rest.fields
import rest.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('account', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='Task',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True, db_index=True)),
                ('modified', models.DateTimeField(auto_now=True)),
                ('started_at', models.DateTimeField(blank=True, default=None, null=True)),
                ('completed_at', models.DateTimeField(blank=True, default=None, null=True)),
                ('stale_after', models.DateTimeField(blank=True, default=None, null=True)),
                ('scheduled_for', models.DateTimeField(blank=True, db_index=True, default=None, null=True)),
                ('cancel_requested', models.BooleanField(blank=True, default=False)),
                ('state', models.IntegerField(choices=[(0, 'scheduled'), (1, 'started'), (2, 'retry_later'), (10, 'completed'), (-1, 'failed'), (-2, 'canceled')], db_index=True, default=0)),
                ('runtime', models.IntegerField(default=0)),
                ('attempts', models.IntegerField(default=0)),
                ('channel', models.CharField(db_index=True, default='tq_task', max_length=200)),
                ('model', models.CharField(db_index=True, max_length=200)),
                ('fname', models.CharField(blank=True, default=None, max_length=200, null=True)),
                ('data', rest.fields.JSONField(blank=True, default=None, null=True)),
                ('reason', models.CharField(blank=True, default=None, max_length=255, null=True)),
            ],
            bases=(models.Model, rest.models.RestModel),
        ),
        migrations.CreateModel(
            name='TaskHook',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True, db_index=True)),
                ('modified', models.DateTimeField(auto_now=True)),
                ('kind', models.CharField(blank=True, max_length=200, null=True)),
                ('data_format', models.CharField(default='json', max_length=32)),
                ('endpoint', models.CharField(blank=True, max_length=200, null=True)),
                ('state', models.IntegerField(choices=[(0, 'inactive'), (1, 'active')], db_index=True, default=1)),
                ('channel', models.CharField(db_index=True, default='tq_hook', max_length=200)),
                ('model', models.CharField(blank=True, db_index=True, max_length=200, null=True)),
                ('group', models.ForeignKey(blank=True, default=None, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='+', to='account.group')),
            ],
            bases=(models.Model, rest.models.RestModel, rest.models.MetaDataModel),
        ),
        migrations.CreateModel(
            name='TaskLog',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True, db_index=True)),
                ('kind', models.CharField(blank=True, default=None, max_length=64, null=True)),
                ('text', models.TextField()),
                ('task', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='logs', to='taskqueue.task')),
            ],
            bases=(models.Model, rest.models.RestModel),
        ),
        migrations.CreateModel(
            name='TaskHookMetaData',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('category', models.CharField(blank=True, db_index=True, default=None, max_length=32, null=True)),
                ('key', models.CharField(db_index=True, max_length=80)),
                ('value_format', models.CharField(max_length=16)),
                ('value', models.TextField()),
                ('int_value', models.IntegerField(blank=True, default=None, null=True)),
                ('float_value', models.IntegerField(blank=True, default=None, null=True)),
                ('parent', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='properties', to='taskqueue.taskhook')),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



try:
    import openpyxl
except Exception:
    openpyxl = None

from .csv import flattenObject, extractFieldNames
from django.http import HttpResponse


def write_fields_to_row(sheet, fields, row=1):
    for col, field in enumerate(fields, start=1):
        sheet.cell(row=row, column=col, value=field)


def qsetToExcel(request, qset, fields, name):
    # Create a new Excel workbook
    workbook = openpyxl.Workbook()

    # Get the active sheet (by default, there's one sheet in a new workbook)
    sheet = workbook.active
    header, field_names = extractFieldNames(fields)

    write_fields_to_row(sheet, header)
    row = 1
    for item in qset:
        row += 1
        write_fields_to_row(sheet, flattenObject(item, field_names), row)

    # Create an HTTP response with the Excel file content
    response = HttpResponse(content_type='application/vnd.openxmlformats-officedocument.spreadsheetml.sheet')
    response['Content-Disposition'] = f'attachment; filename={name}.xlsx'

    # Save the workbook to the response
    workbook.save(response)

    return response




from rest import views as rv
from rest import decorators as rd
from rest import search
from rest import helpers as rh
from rest.uberdict import UberDict
from datetime import datetime, timedelta
from taskqueue import models as tq 

from django.db.models import Sum, Max, Count, Q
from django.db.models.functions import Trunc


@rd.url(r'^task$')
@rd.url(r'^task/(?P<pk>\d+)$')
@rd.login_required
def rest_on_task(request, pk=None):
    return tq.Task.on_rest_request(request, pk)


@rd.url(r'^task/log$')
@rd.url(r'^task/log/(?P<pk>\d+)$')
@rd.login_required
def rest_on_tasklog(request, pk=None):
    return tq.TaskLog.on_rest_request(request, pk)


@rd.url(r'^task/schedule$')
@rd.url(r'^task/schedule/(?P<pk>\d+)$')
@rd.login_required
def rest_on_tasklog(request, pk=None):
    return tq.ScheduledTask.on_rest_request(request, pk)


@rd.urlPOST(r'^task/publish$')
@rd.perm_required("manage_staff")
def rest_on_task_publish(request, pk=None):
    app = request.DATA.get("app").strip()
    module = request.DATA.get("module").strip()
    task_data = request.DATA.get("task_data").strip()
    if not app or not module:
        return rv.restStatus(request, False, error="Both app and module are required.")
    task = tq.Task.Publish(app, module, task_data)
    return rv.restGet(request, task, **tq.Task.getGraph("default"))


@rd.url(r'^task/hook$')
@rd.url(r'^task/hook/(?P<pk>\d+)$')
@rd.login_required
def rest_on_tasklog(request, pk=None):
    return tq.TaskHook.on_rest_request(request, pk)


@rd.url(r'^task/hook/test$')
@rd.perm_required("manage_hooks")
def rest_on_task_hook_test(request):
    hook = request.DATA.get("hook", None)
    if not hook:
        return rv.restStatus(request, False, error="Hook attributes required.")
    data = request.DATA.get("data", None)
    when = request.DATA.get("when", None)
    props = request.DATA.get("props", None)
    category = request.DATA.get("category", None)
    hook = tq.TaskHook(**hook)
    if props:
        hook.save()
        hook.setProperties(props, category=category)
        task = hook.trigger(data, when)
        hook.delete()
    else:
        task = hook.trigger(data, when)
    return rv.restGet(request, task, **tq.Task.getGraph("default"))


@rd.url(r'^restart$')
@rd.perm_required("manage_staff")
def rest_on_restart(request):
    rh.log_print("{} request task engine restart".format(request.member.username))
    tq.Task.RestartEngine()
    return rv.restStatus(request, True)


@rd.url(r'^test$')
@rd.perm_required("manage_staff")
def rest_on_test(request):
    tq.Task.PublishTest(request.DATA.get("test_count", 1, field_type=int), request.DATA.get("sleep_time", 10.0, field_type=float))
    return rv.restStatus(request, True)


@rd.url(r'^task/status$')
def rest_on_task_status(request):
    out = UberDict()
    last_completed = tq.Task.objects.filter(state=tq.TASK_STATE_COMPLETED).last()
    if last_completed:
        out.last_completed = last_completed.completed_at
    last_scheduled = tq.Task.objects.all().last()
    if last_scheduled:
        out.last_scheduled = last_scheduled.created
    stale = datetime.now() - timedelta(minutes=request.DATA.get("minutes_back", 60))
    qset = tq.Task.objects.filter(modified__gte=stale)
    status = qset.aggregate(
        retry=Count('state', filter=Q(state=tq.TASK_STATE_RETRY)),
        completed=Count('state', filter=Q(state=tq.TASK_STATE_COMPLETED)),
        failed=Count('state', filter=Q(state=tq.TASK_STATE_FAILED)),
        running=Count('state', filter=Q(state=tq.TASK_STATE_STARTED)),
        backlog=Count('state', filter=Q(state=tq.TASK_STATE_SCHEDULED)))
    status["total"] = qset.count()
    out.update(status)
    return rv.restGet(request, out)


@rd.urlGET('workers')
def rest_on_runners(request):
    workers = tq.TaskWorkerClient.GET_ONLINE()
    if request.DATA.get("graph") == "detailed":
        output = []
        for r in workers:
            stats = tq.TaskWorkerClient.GET_STATS(r)
            stats.id = r
            output.append(stats)
        workers = output
    return rv.restReturn(request, dict(data=workers))


# write a django query to 

@rd.url(r'^task/stats$')
def rest_on_stats(request):
    now = datetime.now() 
    when = now - timedelta(days=request.DATA.get("days", 7, field_type=int))
    qset = tq.Task.objects.filter(
        created__gte=when)
    reports = list(
        qset
        .annotate(day=Trunc('created', 'day'))
        .values('day')
        .annotate(completed=Count('state', filter=Q(state=10)))
        .annotate(failed=Count('state', filter=Q(state=-1)))
        .annotate(longest=Max('runtime')))
    output = []
    
    when = when.replace(hour=0, minute=0, second=0)
    while when < now:
        dr = None
        for r in reports:
            day = r["day"]
            if day.day == when.day and day.month == when.month:
                dr = r
                break
        if dr is not None:
            output.append(dr)
        else:
            output.append(dict(day=when, completed=0, failed=0, longest=0))
        when += timedelta(days=1)

    status = qset.aggregate(
        retry=Count('state', filter=Q(state=tq.TASK_STATE_RETRY)),
        completed=Count('state', filter=Q(state=tq.TASK_STATE_COMPLETED)),
        failed=Count('state', filter=Q(state=tq.TASK_STATE_FAILED)),
        running=Count('state', filter=Q(state=tq.TASK_STATE_STARTED)),
        backlog=Count('state', filter=Q(state=tq.TASK_STATE_SCHEDULED)))
    return rv.restGet(request, dict(stats=output, status=status))


@rd.url(r'^test/webrequest$')
def rest_on_webrequest(request):
    msg = "Successfully received the following data\n{}".format(str(request.DATA.toDict()))
    return rv.restStatus(request, True, msg=msg)



from django.contrib import admin
from .models import *


class GroupAdmin(admin.ModelAdmin):
    search_fields = ["name"]
    list_display = ('name', 'uuid', 'kind', 'parent')
    fields = ('name', 'short_name', 'kind', 'parent', 'is_active')


class MemberAdmin(admin.ModelAdmin):
    date_hierarchy = "date_joined"
    list_display = ('username', 'email', 'first_name', 'last_name', 'last_login',)
    search_fields = ['email', 'first_name', 'last_name', 'username']
    fields = ('display_name', 'username', 'first_name', 'last_name', 'email', 'is_staff','is_superuser', 'is_active',)


class MembershipAdmin(admin.ModelAdmin):
    raw_id_fields = ['member', 'group']
    date_hierarchy = "created"
    list_display = ('member', 'role', 'group', 'created', 'state',)
    search_fields = ['role', 'member__email', 'member__last_name', 'member__first_name']


class InviteList(admin.ModelAdmin):
    date_hierarchy = "created"
    list_display = ('email', 'name', 'created', 'invited_by', 'approved')
    fields = ('name', 'approved')
    search_fields = ['email', 'name']


class AuthTokenAdmin(admin.ModelAdmin):
    raw_id_fields = ['member']
    date_hierarchy = "created"
    list_display = ('created', 'member')


class MemberDeviceAdmin(admin.ModelAdmin):
    raw_id_fields = ['member']
    date_hierarchy = "created"
    list_display = ('created', 'member', "uuid", "kind")


class SessionAdmin(admin.ModelAdmin):
    raw_id_fields = ['member']
    date_hierarchy = "created"
    list_display = ('created', "last_activity", 'member', "ip", "device")


admin.site.register(Member, MemberAdmin)
admin.site.register(Group, GroupAdmin)
admin.site.register(Membership, MembershipAdmin)
admin.site.register(AuthToken, AuthTokenAdmin)
admin.site.register(MemberDevice, MemberDeviceAdmin)
admin.site.register(AuthSession, SessionAdmin)





import base64
import re
import binascii
import string
import hmac
from hashlib import md5, sha512, sha256
from Crypto.Random import random as crypt_random
from hashids import Hashids


def generateKey(bit_size=128, allow_punctuation=False):
    byte_size = int(bit_size / 8)
    return randomString(byte_size, allow_punctuation)


def generateUUID(*args, **kwargs):
    upper = kwargs.get("upper", True)
    max_length = kwargs.get("max_length", None)
    uuid = ""
    for key in args:
        if isinstance(key, float):
            key = str(float)
        if isinstance(key, int):
            uuid += Hashids().encrypt(key)
        if isinstance(key, str):
            uuid += toString(md5(toBytes(key)).hexdigest())
    if len(uuid) > 125:
        uuid = uuid[:125]
    if max_length != None:
        uuid = uuid[:max_length]
    if upper:
        return uuid.upper()
    return uuid


def get_random_bits(bit_size=128):
    return crypt_random.getrandbits(bit_size)


def randomString(str_size=128, allow_punctuation=False, char_list=None):
    if char_list is None:
        char_list = string.ascii_letters + string.digits
    if allow_punctuation:
        char_list += string.punctuation
    return ''.join([crypt_random.choice(char_list) for n in range(str_size)])


def randomCode(str_size=6):
    return randomString(str_size, False, string.digits)


def hash(data):
    return hashMD5(data)


def hashit(data, salt=None):
    return hashSHA512(data, salt)


def obfuscateID(label, int_id):
    return Hashids(label).encode(int_id)


def unobfuscateID(label, obfuscated_id):
    res = Hashids(label).decode(obfuscated_id)
    if len(res):
        return res[0]
    return None


def toString(value):
    if isinstance(value, bytes):
        value = value.decode()
    elif isinstance(value, bytearray):
        value = value.decode("utf-8")
    elif isinstance(value, (int, float)):
        value = str(value)
    return value


def toBytes(value):
    if isinstance(value, str):
        value = value.encode("utf-8")
    elif isinstance(value, bytearray):
        value = bytes(value)
    return value


def toByteArray(value):
    if isinstance(value, bytearray):
        return value
    elif isinstance(value, str):
        return bytearray(value, 'utf-8')
    elif isinstance(value, bytes):
        return bytearray(value)
    return value


def toHex(value):
    return toBytes(value).hex().upper()


def hexToByteArray(value):
    return bytearray.fromhex(value)


def hexToString(value):
    return bytes.fromhex(value).decode('utf-8')


def toBase64(value, as_str=True):
    if isinstance(value, str):
        value = toBytes(value)
    if as_str:
        return toString(base64.b64encode(value))
    return base64.b64encode(value)


def fromBase64(value, as_str=True):
    if as_str:
        return toString(base64.b64decode(value))
    return base64.b64decode(value)


def hashSHA256(data, salt=None):
    bdat = toBytes(data)
    if salt is not None:
        bdat = toBytes(salt) + bdat
    return toString(sha256(bdat).hexdigest())


def hashSHA512(data, salt=None):
    bdat = toBytes(data)
    if salt is not None:
        bdat = toBytes(salt) + bdat
    return toString(sha512(toBytes(bdat)).hexdigest())


def hashMD5(data):
    return toString(md5(toBytes(data)).hexdigest())


def getSSHSignature(pub_key):
    data = pub_key.strip()
    # accept either base64 encoded data or full pub key file,
    # same as `fingerprint_from_ssh_pub_key
    if (re.search(r'^ssh-(?:rsa|dss|ed25519) ', data)):
        data = data.split(None, 2)[1]
    # Python 2/3 hack. May be a better solution but this works.
    data = toBytes(data)
    digest = sha256(binascii.a2b_base64(data)).digest()
    encoded = toString(base64.b64encode(digest).rstrip(b'='))  # ssh-keygen strips this
    return "SHA256:" + encoded


def pad(data, block_size=16):
    # PKCS7Padding
    # if isinstance(data, str):
    #     data = str(data)
    length = block_size - (len(data) % block_size)
    return data + (chr(length) * length).encode()


def unpad(data):
    # PKCS7Padding
    return data[:-(data[-1] if type(data[-1]) == int else ord(data[-1]))]

def generateHMAC(label, val, length, encoding="base32", secret=None):
    if secret is None:
        secret = hashSHA256(label)
    ret = hmac.new(toBytes(secret + label), toBytes(val), sha512)
    if encoding == "base32":
        retstr = toString(base64.b32encode(ret.digest()))
    else:
        retstr = toHex(ret.digest())
    return retstr[:length].lower()



from django.db import models
from django.conf import settings

from rest import models as rm
from rest import helpers as rh
import re

UNDEFINED = "TYUIFGHJ"

"""
Rule defines a set of checks that will create an incident from an incoming event.

There is no subscriptions.  Group with kind incident are created as buckets.
When a rule is triggered it will create a new incident

the incident will be assigned to the rule.group
the incident priority will be assigned to the rule.priority

Events can be bundled into a single Incident by setting bundle=<minutes>.  
So if bundle is set to 60 then for 60 minutes after the first event all events will get 
added to to same incident.

"""


class Rule(models.Model, rm.RestModel):
    class RestMeta:
        SEARCH_FIELDS = ["name", "group__name"]
        CAN_DELETE = True
        # VIEW_PERMS = ["example_permission"]
        GRAPHS = {
            "default": {
                "graphs": {
                    "group": "basic",
                    "created_by": "basic"
                },
            },
            "basic": {
                "fields": [
                    "created",
                    "priority",
                    "name",
                    "category",
                    "priority",
                    "action",
                    "action_after",
                    "bundle",
                    "bundle_by"
                ]
            }
        }

    created = models.DateTimeField(auto_now_add=True)
    modified = models.DateTimeField(auto_now=True)

    name = models.CharField(max_length=200)
    # the group the rule gets assigned to when triggered
    group = models.ForeignKey("account.Group", on_delete=models.CASCADE, null=True, default=None)
    # category allows us to limit running rules to only those with a category
    # if category is null then this will run on all events?
    category = models.CharField(max_length=200, db_index=True)
    created_by = models.ForeignKey("account.Member", on_delete=models.CASCADE, null=True, default=None)
    # this specifies the order the rule runs in and will assign the priority to the created incident
    priority = models.IntegerField(default=10, db_index=True)  # 1-10, 1 being the highest
    # if multiple events fire the same rule it will just bundle them together
    # this is in minutes (default=0=OFF)
    bundle = models.IntegerField(default=0)  # 0=off
    bundle_by = models.IntegerField(default=3)  # 0=none, 1=hostname, 2=component, 3=component and hostname
    # an action to perform if this rule is hit, default is notify
    # examples (ignore, notify, task:app_name:fname:channel)
    action = models.CharField(max_length=200, default=None, null=True)
    action_after = models.IntegerField(default=0)  # how many incidents before firing action
    
    def run(self, event):
        for check in self.checks.all():
            if not check.run(event):
                return False
        return True


class RuleCheck(models.Model, rm.RestModel):
    class RestMeta:
        SEARCH_FIELDS = ["name"]
        CAN_DELETE = True
        # VIEW_PERMS = ["example_permission"]
        GRAPHS = {
            "default": {
                "graphs": {
                    "parent": "basic"
                },
            }
        }

    created = models.DateTimeField(auto_now_add=True)
    modified = models.DateTimeField(auto_now=True)
    parent = models.ForeignKey(Rule, on_delete=models.CASCADE, related_name="checks")
    # the lower the number the sooner it runs
    name = models.CharField(max_length=200)
    # the order for which this check runs
    index = models.IntegerField(default=0)
    comparator = models.CharField(max_length=32, default="==")
    field_name = models.CharField(max_length=124, default=None, null=True)
    value = models.CharField(max_length=124, default="")
    value_type = models.IntegerField(default=0)  # 0=int 1=float 2=string
 
    def run(self, event):
        # TODO ians look into does this support pulling metadata.....
        field_value = event.getProperty(self.field_name, UNDEFINED)
        if field_value == UNDEFINED:
            return False
        comp_value = self.value
        try:
            if self.comparator != "contains":
                if self.value_type == 0:
                    comp_value = int(comp_value)
                    field_value = int(field_value)
                elif self.value_type == 1:
                    comp_value = float(comp_value)
                    field_value = float(field_value)
        except Exception:
            return False

        if self.comparator in ["eq", "=="]:
            return field_value == comp_value
        if self.comparator == ">":
            return field_value > comp_value
        if self.comparator == ">=":
            return field_value >= comp_value        
        if self.comparator == "<":
            return field_value < comp_value     
        if self.comparator == "<=":
            return field_value <= comp_value
        if self.comparator == "contains":
            field_value = str(field_value)
            comp_value = str(comp_value)
            return field_value.find(comp_value) != -1
        if self.comparator == "regex":
            return re.match(comp_value, field_value)
        return False    




from rest import settings
from rest import datem as date_util
from django.template.defaultfilters import slugify as django_slugify
from collections import OrderedDict
from datetime import datetime, timedelta

app_settings = settings.getAppSettings("metrics")
GRANULARITIES = ['seconds', 'minutes', 'hourly', 'daily', 'weekly', 'monthly', 'yearly']


def getLocalDate(date, group=None):
    if date is None:
        date = datetime.utcnow()
    elif isinstance(date, str):
        date = date_util.parseDateTime(date)

    if group is not None:
        date = group.getLocalTime(date)
    elif settings.METRICS_TIMEZONE:
        date = date_util.convertToLocalTime(settings.METRICS_TIMEZONE, date)
    return date


def to_string(value):
    if isinstance(value, bytes):
        value = value.decode()
    elif isinstance(value, bytearray):
        value = value.decode("utf-8")
    elif isinstance(value, (int, float)):
        value = str(value)
    return value


def to_int_list(values):
    """Converts the given list of vlues into a list of integers. If the
    integer conversion fails (e.g. non-numeric strings or None-values), this
    filter will include a 0 instead."""
    results = []
    for v in values:
        try:
            results.append(int(v))
        except (TypeError, ValueError):
            results.append(0)
    return results


def to_int(value):
    """Converts the given string value into an integer. Returns 0 if the
    conversion fails."""
    try:
        return int(value)
    except (TypeError, ValueError):
        return 0


def strip_metric_prefix(value):
    """Strips the "m:<slug>" prefix from a metric.
    Applying this filter to the keys for each metric will have the following
    results:

    * Seconds -- from: ``m:<slug>:s:<yyyy-mm-dd-hh-MM-SS>`` to ``<yyyy-mm-dd-hh-MM-SS>``
    * Minutes -- from: ``m:<slug>:i:<yyyy-mm-dd-hh-MM>`` to ``<yyyy-mm-dd-hh-MM>``
    * Hourly -- from: ``m:<slug>:h:<yyyy-mm-dd-hh>`` to ``<yyyy-mm-dd-hh>``
    * Daily -- from: ``m:<slug>:<yyyy-mm-dd>`` to ``<yyyy-mm-dd>``
    * Weekly -- from ``m:<slug>:w:<num>`` to ``w:<num>``
    * Monthly -- from ``m:<slug>:m:<yyyy-mm>`` to ``m:<yyyy-mm>``
    * Yearly -- from ``m:<slug>:y:<yyyy>`` to ``y:<yyyy>``

    """
    return ':'.join(value.split(":")[2:])


def metric_slug(value):
    """Given a redis key value for a metric, returns only the slug.
    Applying this filter to the keys for each metric will have the following
    results:

    * Converts ``m:foo:s:<yyyy-mm-dd-hh-MM-SS>`` to ``foo``
    * Converts ``m:foo:i:<yyyy-mm-dd-hh-MM>`` to ``foo``
    * Converts ``m:foo:h:<yyyy-mm-dd-hh>`` to ``foo``
    * Converts ``m:foo:<yyyy-mm-dd>`` to ``foo``
    * Converts ``m:foo:w:<num>`` to ``foo``
    * Converts ``m:foo:m:<yyyy-mm>`` to ``foo``
    * Converts ``m:foo:y:<yyyy>`` to ``foo``

    """
    return value.split(":")[1]


def granularities(min_granularity=None, max_granularity=None):
    """Returns a generator of all possible granularities based on the
    MIN_GRANULARITY and MAX_GRANULARITY settings.
    """
    if min_granularity is None:
        min_granularity = app_settings.MIN_GRANULARITY
    if max_granularity is None:
        max_granularity = app_settings.MAX_GRANULARITY
    keep = False
    for g in GRANULARITIES:
        if g == min_granularity and not keep:
            keep = True
        elif g == max_granularity and keep:
            keep = False
        if keep:
            if g == max_granularity:
                keep = False
            yield g


def get_metric_key_pattern(granularity, slug, date):
    """ The Redis metric key and date formatting patterns for each key, by granularity"""
    patterns_by_granularity = {
        "seconds": {"key": "m:{0}:s:{1}", "date_format": "%Y-%m-%d-%H-%M-%S"},
        "minutes": {"key": "m:{0}:i:{1}", "date_format": "%Y-%m-%d-%H-%M"},
        "hourly": {"key": "m:{0}:h:{1}", "date_format": "%Y-%m-%d-%H"},
        "daily": {"key": "m:{0}:{1}", "date_format": "%Y-%m-%d"},
        "weekly": {
            "key": "m:{0}:w:{1}",
            "date_format": get_weekly_date_format(date),
        },
        "monthly": {"key": "m:{0}:m:{1}", "date_format": "%Y-%m"},
        "yearly": {"key": "m:{0}:y:{1}", "date_format": "%Y"},
    }
    pattern_for_granularity = patterns_by_granularity[granularity]
    fmt = pattern_for_granularity["date_format"]
    date_string = date.strftime(fmt)
    return pattern_for_granularity["key"].format(slug, date_string)


def get_weekly_date_format(date):
    if app_settings.USE_ISO_WEEK_NUMBER:
        # We can return instantly because ISO week start on monday
        return "{year}-{week_no}".format(year=date.isocalendar()[0], week_no=date.isocalendar()[1])
    return "%Y-%{0}".format('W' if app_settings.MONDAY_FIRST_DAY_OF_WEEK else 'U')


def get_weekly_date(date):
    return date.replace(hour=0, minute=0, second=0, microsecond=0) - timedelta(days=date.weekday())


def build_key_patterns(slug, date, min_granularity=None, max_granularity=None):
    """Builds an OrderedDict of metric keys and patterns for the given slug
    and date."""
    # we want to keep the order, from smallest to largest granularity
    patts = OrderedDict()
    for g in granularities(min_granularity=min_granularity, max_granularity=max_granularity):
        patts[g] = get_metric_key_pattern(g, slug, date)
    return patts


def slugify(slug):
    return slug.replace(" ", "_").replace(":", "|").lower()


def build_keys(slug, date=None, granularity='all', min_granularity=None, max_granularity=None):
    """Builds redis keys used to store metrics.

    * ``slug`` -- a slug used for a metric, e.g. "user-signups"
    * ``date`` -- (optional) A ``datetime.datetime`` object used to
      generate the time period for the metric. If omitted, the current date
      and time (in UTC) will be used.
    * ``granularity`` -- Must be one of: "all" (default), "yearly",
    "monthly", "weekly", "daily", "hourly", "minutes", or "seconds".

    Returns a list of strings.

    """
    slug = slugify(slug)  # Ensure slugs have a consistent format
    if date is None:
        date = getLocalDate(datetime.utcnow())
    if min_granularity is None and granularity != "all":
        min_granularity = granularity
    patts = build_key_patterns(slug, date, min_granularity=min_granularity, max_granularity=max_granularity)
    if granularity == "all":
        return list(patts.values())
    return [patts[granularity]]

def slug_to_label(slug):
    lbl = strip_metric_prefix(slug)
    if lbl.startswith("h:"):
        # hourly
        return f"{lbl[-2:]}:00"
    elif lbl.startswith("m:") or lbl.startswith("y:"):
        return lbl[2:]
    return lbl

def date_for_granulatiry(date, granularity):
    if granularity == "minutes":
        return date.replace(second=0, microsecond=0)
    elif granularity == "hourly":
        return date.replace(minute=0, second=0, microsecond=0)
    elif granularity == "daily":
        return date.replace(hour=0, minute=0, second=0, microsecond=0)
    elif granularity == "weekly":
        return get_weekly_date(date)
    elif granularity == "monthly":
        return date.replace(day=1, hour=0, minute=0, second=0, microsecond=0)
    return date.replace(month=1, day=1, hour=0, minute=0, second=0, microsecond=0)


def start_by_granularity(granularity, date=None, samples=None):
    if date is None:
        date = getLocalDate(date)
    if granularity == "minutes":
        if samples is None:
            samples = 15
        return date.replace(second=0, microsecond=0) - timedelta(minutes=samples)
    elif granularity == "hourly":
        if samples is None:
            samples = 25
        return date.replace(minute=0, second=0, microsecond=0) - timedelta(hours=samples)
    elif granularity == "daily":
        if samples is None:
            samples = 7
        return date.replace(hour=0, minute=0, second=0, microsecond=0) - timedelta(days=samples)
    elif granularity == "weekly":
        if samples is None:
            samples = 8
        return get_weekly_date(date - timedelta(days=samples*7))
    elif granularity == "monthly":
        if samples is None:
            samples = 8
        return (date - timedelta(days=30*samples)).replace(day=1, hour=0, minute=0, second=0, microsecond=0)
    elif granularity == "yearly":
        if samples is None:
            samples = 6
        return (date - timedelta(days=366*samples)).replace(day=1, hour=0, minute=0, second=0, microsecond=0)

    if samples is None:
        samples = 7
    return date - timedelta(days=samples)


def date_range(granularity, since, to=None):
    """Returns a generator that yields ``datetime.datetime`` objects from
    the ``since`` date until ``to`` (default: *now*).

    * ``granularity`` -- The granularity at which the generated datetime
      objects should be created: seconds, minutes, hourly, daily, weekly,
      monthly, or yearly
    * ``since`` -- a ``datetime.datetime`` object, from which we start
      generating periods of time. This can also be ``None``, and will
      default to the past 7 days if that's the case.
    * ``to`` -- a ``datetime.datetime`` object, from which we start
      generating periods of time. This can also be ``None``, and will
      default to now if that's the case.

    If ``granularity`` is one of daily, weekly, monthly, or yearly, this
    function gives objects at the daily level.

    If ``granularity`` is one of the following, the number of datetime
    objects returned is capped, otherwise this code is really slow and
    probably generates more data than we want:

        * hourly: returns at most 720 values (~30 days)
        * minutes: returns at most 480 values (8 hours)
        * second: returns at most 300 values (5 minutes)

    For example, if granularity is "seconds", we'll receive datetime
    objects that differ by 1 second each.

    """
    if since is None:
        since = start_by_granularity(granularity)

    if to is None:
        to = getLocalDate(datetime.utcnow())
    elapsed = (to - since)

    # Figure out how many units to generate for the elapsed time.
    # I'm going to use `granularity` as a keyword parameter to timedelta,
    # so I need to change the wording for hours and anything > days.
    if granularity == "seconds":
        units = elapsed.total_seconds()
        units = 300 if units > 300 else units
    elif granularity == "minutes":
        units = elapsed.total_seconds() / 60
        units = 480 if units > 480 else units
    elif granularity == "hourly":
        granularity = "hours"
        units = elapsed.total_seconds() / 3600
        units = 720 if units > 720 else units
    elif granularity == "daily":
        granularity = "days"
        units = elapsed.days + 1
    elif granularity == "weekly":
        output = []
        start = get_weekly_date(since)
        while start < to:
            output.append(start)
            start += timedelta(days=7)
        return output
    elif granularity == "monthly":
        granularity = "days"
        months = diff_month(to, since)
        output = []
        for i in range(months):
            output.append(to.replace(day=1, hour=0, minute=0, second=0, microsecond=0))
            to = to.replace(day=10) - timedelta(days=31)
        return output
    elif granularity == "yearly":
        output = []
        start = since + timedelta(days=1)
        start = start.replace(month=1, day=1, hour=0, minute=0, second=0, microsecond=0)
        end = to + timedelta(days=365)
        while start < end:
            output.append(start)
            start += timedelta(days=368)
            start = start.replace(month=1, day=1, hour=0, minute=0, second=0, microsecond=0)
        output.reverse()
        return output

    return (to - timedelta(**{granularity: u}) for u in range(int(units)))


def diff_month(d1, d2):
    return (d1.year - d2.year) * 12 + d1.month - d2.month



from django.db import models
from rest import models as rm
import uuid


# DEPRECATED!
class Permission(models.Model):
    membership = models.ForeignKey("account.Membership", related_name="permissions", on_delete=models.CASCADE)
    name = models.CharField(max_length=255)


class AuthAccount(models.Model, rm.RestModel):
    class RestMeta:
        SEARCH_FIELDS = ["member__first_name", "member__last_name"]
        NO_SHOW_FIELDS = ["pin", "pan"]
        GRAPHS = {
            "default": {
                "graphs": {
                    "member": "basic",
                },
                "extra": [
                    "pan_last_4",
                ],
            },
            "list": {
                "graphs": {
                    "self": "default"
                }
            },
            "detailed": {
                "graphs": {
                    "self": "default"
                }
            }
        }

    created = models.DateTimeField(auto_now_add=True, editable=False)
    pan = models.TextField(db_index=True)
    pin = models.CharField(max_length=64, blank=True, null=True, default=None)
    kind = models.CharField(max_length=128, blank=True, null=True, default=None)
    member = models.ForeignKey("account.Member", related_name="auth_accounts", on_delete=models.CASCADE)
    state = models.IntegerField(default=1, choices=[(0, "disabled"), (1, "enabled")], db_index=True)

    @property
    def pan_last_4(self):
        if not self.pan:
            return ""
        return self.pan[-4:]

    def set_kind(self, kind):
        self.kind = kind
        if kind == "authtoken":
            request = self.getActiveRequest()
            if bool(request):
                pan = request.DATA.get("pan")
                if not pan or len(pan) < 8:
                    # do not let user set pans on auth tokens
                    self.pan = str(uuid.uuid1())
                    # override it in the data to make sure
                    request.DATA.set("pan", self.pan)

    @classmethod
    def queryFromRequest(cls, request, qset):
        if not request.user.is_staff:
            member_id = request.DATA.get(["member", "member_id"])
            if not member_id:
                member_id = request.member.id
            qset = qset.filter(member_id=member_id)
        return super(AuthAccount, cls).queryFromRequest(request, qset)

    @classmethod
    def on_rest_list_filter(cls, request, qset=None):
        if request.group:            
            members = request.group.getMembers(as_member=True, include_parents=False)
            qset = qset.filter(member__in=members)
        return qset

    def __unicode__(self):
        return "{o.member} <{o.pan}>".format(o=self)



# Generated by Django 4.1.3 on 2022-12-17 02:50

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('incident', '0004_serverossecalert_ssh_sig_alter_rulecheck_parent'),
    ]

    operations = [
        migrations.AddField(
            model_name='incident',
            name='component',
            field=models.CharField(db_index=True, default=None, max_length=200, null=True),
        ),
        migrations.AlterField(
            model_name='incident',
            name='state',
            field=models.IntegerField(choices=[(0, 'new'), (1, 'opened'), (2, 'paused'), (3, 'ignored'), (4, 'resolved')], default=0),
        ),
    ]
